/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA, S3 GRAPHICS, AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifndef _VIA_BIOS_H
#define _VIA_BIOS_H

#define     VIA_BIOS_REG_TABLE_MAX_NUM      32
#define     VIA_BIOS_REG_LCD_MAX_NUM        48
#define     VIA_BIOS_NUM_RES                17
#define     VIA_BIOS_NUM_REFRESH            5
#define     VIA_BIOS_NUM_LCD_SUPPORT_MASK   16
#define     VIA_BIOS_NUM_LCD_POWER_SEQ      4
#define     VIA_BIOS_NUM_PANEL              7
#define     VIA_BIOS_MAX_NUM_MPATCH2        18
#define     VIA_BIOS_MAX_NUM_MPATCH1        9
#define     VIA_BIOS_MAX_NUM_CTREXP         5
#define     VIA_BIOS_MAX_NUM_TV_REG         144		/* 00 - 8F, tv2,tv3,ch7019 are the same */
#define     VIA_BIOS_MAX_NUM_SAA7108_TV_REG 176		/* 00 - AF */
#define     VIA_BIOS_MAX_NUM_TV_CRTC        32
#define     VIA_BIOS_NUM_TV_SPECIAL_REG     8
#define     VIA_BIOS_MAX_NUM_TV_PATCH       8
#define     VIA_BIOS_NUM_TV_OTHER           16
#define     VIA_BIOS_NUM_TV2                2
#define     VIA_BIOS_NUM_TV3                6
#define     VIA_BIOS_NUM_SAA7108            4
#define     VIA_BIOS_NUM_CH7019	     3

#define     VIA_RES_640X480                 0
#define     VIA_RES_800X600                 1
#define     VIA_RES_1024X768                2
#define     VIA_RES_1152X864                3
#define     VIA_RES_1280X1024               4
#define     VIA_RES_1600X1200               5
#define     VIA_RES_1440X1050               6
#define     VIA_RES_1280X768                7
#define     VIA_RES_1280X960                8
#define     VIA_RES_1920X1440               9
#define     VIA_RES_848X480                 10
#define     VIA_RES_1400X1050               11
#define     VIA_RES_720X480                 12
#define     VIA_RES_720X576                 13
#define     VIA_RES_1024X512                14
#define     VIA_RES_856X480                 15
#define     VIA_RES_1024X576                16
#define     VIA_RES_INVALID                 255

#define     VIA_TVRES_640X480               0
#define     VIA_TVRES_800X600               1
#define     VIA_TVRES_1024X768              2
#define     VIA_TVRES_848X480               3
#define     VIA_TVRES_720X480               4
#define     VIA_TVRES_720X576               5

#define     TVTYPE_NONE                     0x00
#define     TVTYPE_NTSC                     0x01
#define     TVTYPE_PAL                      0x02

#define     TVOUTPUT_NONE                   0x00
#define     TVOUTPUT_COMPOSITE              0x01
#define     TVOUTPUT_SVIDEO                 0x02
#define     TVOUTPUT_RGB                    0x04
#define     TVOUTPUT_YCBCR                  0x08
#define     TVOUTPUT_SC                     0x16

#define     VIA_NONETV                      0
#define     VIA_TV2PLUS                     1
#define     VIA_TV3                         2
#define     VIA_CH7009	                     3
#define     VIA_CH7019                      4
#define     VIA_SAA7108                     5
#define     VIA_CH7005                      6
#define     VIA_VT1622A                     7

#define     VIA_TVNORMAL                    0
#define     VIA_TVOVER                      1


typedef struct _VIABIOSSTDVGATABLE {
    unsigned char           columns;
    unsigned char           rows;
    unsigned char           fontHeight;
    unsigned int            pageSize;
    unsigned char           SR[5];
    unsigned char           misc;
    unsigned char           CR[25];
    unsigned char           AR[20];
    unsigned char           GR[9];
} VIABIOSStdVGATableRec, *VIABIOSStdVGATablePtr;


typedef struct _VIABIOSREFRESHTABLE {
    unsigned char           refresh;
    unsigned int            VClk;
    unsigned char           CR[14];
} VIABIOSRefreshTableRec, *VIABIOSRefreshTablePtr;


typedef struct _VIABIOSREGTABLE {
    unsigned char   port[VIA_BIOS_REG_TABLE_MAX_NUM];
    unsigned char   offset[VIA_BIOS_REG_TABLE_MAX_NUM];
    unsigned char   mask[VIA_BIOS_REG_TABLE_MAX_NUM];
    unsigned char   data[VIA_BIOS_REG_TABLE_MAX_NUM];
    int     numEntry;
} VIABIOSRegTableRec, *VIABIOSRegTablePtr;


typedef struct _VIAVMODEENTRY {
    unsigned short          Width;
    unsigned short          Height;
    unsigned short          Bpp;
    unsigned short          Mode;
    unsigned short          MemNeed;
    unsigned short          MClk;
    unsigned short          VClk;
    VIABIOSStdVGATableRec   stdVgaTable;
    VIABIOSRegTableRec      extModeExtTable;
} VIAModeEntry, *VIAModeEntryPtr;


typedef struct _VIALCDMODEENTRY {
    unsigned int    LCDClk;
    unsigned int    VClk;
    unsigned int    LCDClk_12Bit;
    unsigned int    VClk_12Bit;
    unsigned char   port[VIA_BIOS_REG_LCD_MAX_NUM];
    unsigned char   offset[VIA_BIOS_REG_LCD_MAX_NUM];
    unsigned char   data[VIA_BIOS_REG_LCD_MAX_NUM];
    int     numEntry;
} VIALCDModeEntry, *VIALCDModeEntryPtr;


typedef struct _VIALCDMPATCHENTRY {
    unsigned char   Mode;
    unsigned int    LCDClk;
    unsigned int    VClk;
    unsigned int    LCDClk_12Bit;
    unsigned int    VClk_12Bit;
    unsigned char   port[VIA_BIOS_REG_LCD_MAX_NUM];
    unsigned char   offset[VIA_BIOS_REG_LCD_MAX_NUM];
    unsigned char   data[VIA_BIOS_REG_LCD_MAX_NUM];
    int     numEntry;
} VIALCDMPatchEntry, *VIALCDMPatchEntryPtr;


typedef struct _VIALCDMODEFIX {
    unsigned char   reqMode[32];
    unsigned char   fixMode[32];
    int     numEntry;
} VIALCDModeFixRec, *VIALCDModeFixRecPtr;


typedef struct _VIALCDPOWERSEQUENCE {
    unsigned char   powerSeq;
    unsigned char   port[4];
    unsigned char   offset[4];
    unsigned char   mask[4];
    unsigned char   data[4];
    unsigned int    delay[4];
    int     numEntry;
} VIALCDPowerSeqRec, *VIALCDPowerSeqRecPtr;


typedef struct _VIALCDMODETABLE {
    unsigned char                   fpIndex;
    unsigned char                   fpSize;
    unsigned char                   powerSeq;
    int                     numMPatchDP2Ctr;
    int                     numMPatchDP2Exp;
    int                     numMPatchDP1Ctr;
    int                     numMPatchDP1Exp;
    unsigned int                    SuptMode[VIA_BIOS_NUM_LCD_SUPPORT_MASK];
    VIALCDModeEntry         FPconfigTb;
    VIALCDModeEntry         InitTb;
    VIALCDMPatchEntry       MPatchDP2Ctr[VIA_BIOS_MAX_NUM_MPATCH2];
    VIALCDMPatchEntry       MPatchDP2Exp[VIA_BIOS_MAX_NUM_MPATCH2];
    VIALCDMPatchEntry       MPatchDP1Ctr[VIA_BIOS_MAX_NUM_MPATCH1];
    VIALCDMPatchEntry       MPatchDP1Exp[VIA_BIOS_MAX_NUM_MPATCH1];
    VIALCDModeEntry         LowResCtr;
    VIALCDModeEntry         LowResExp;
    VIALCDModeEntry         MCtr[VIA_BIOS_MAX_NUM_CTREXP];
    VIALCDModeEntry         MExp[VIA_BIOS_MAX_NUM_CTREXP];
} VIALCDModeTableRec, *VIALCDModePtr;


typedef struct _VIATVMASKTABLE {
    unsigned char   TV[VIA_BIOS_MAX_NUM_TV_REG];
    unsigned char   CRTC1[VIA_BIOS_MAX_NUM_TV_CRTC];
    unsigned char   CRTC2[VIA_BIOS_MAX_NUM_TV_CRTC];
    unsigned char   misc1;
    unsigned char   misc2;
    int     numTV;
    int     numCRTC1;
    int     numCRTC2;
} VIABIOSTVMASKTableRec, *VIABIOSTVMASKTablePtr;


typedef struct _VIATV3MODETABLE {
    unsigned char   TVNTSC[VIA_BIOS_MAX_NUM_TV_REG];
    unsigned char   CRTCNTSC1[VIA_BIOS_MAX_NUM_TV_CRTC];
    unsigned char   MiscNTSC1[VIA_BIOS_NUM_TV_SPECIAL_REG];
    unsigned char   MiscNTSC2[VIA_BIOS_NUM_TV_SPECIAL_REG];
    unsigned char   CRTCNTSC2_8BPP[VIA_BIOS_MAX_NUM_TV_CRTC];
    unsigned char   CRTCNTSC2_16BPP[VIA_BIOS_MAX_NUM_TV_CRTC];
    unsigned char   CRTCNTSC2_32BPP[VIA_BIOS_MAX_NUM_TV_CRTC];
    unsigned int    PatchNTSC2[VIA_BIOS_MAX_NUM_TV_PATCH];
    unsigned int    RGBNTSC[VIA_BIOS_NUM_TV_OTHER];
    unsigned int    YCbCrNTSC[VIA_BIOS_NUM_TV_OTHER];
    unsigned int    SDTV_RGBNTSC[VIA_BIOS_NUM_TV_OTHER];
    unsigned int    SDTV_YCbCrNTSC[VIA_BIOS_NUM_TV_OTHER];
    unsigned int    DotCrawlNTSC[VIA_BIOS_NUM_TV_OTHER];
    unsigned char   TVPAL[VIA_BIOS_MAX_NUM_TV_REG];
    unsigned char   CRTCPAL1[VIA_BIOS_MAX_NUM_TV_CRTC];
    unsigned char   MiscPAL1[VIA_BIOS_NUM_TV_SPECIAL_REG];
    unsigned char   MiscPAL2[VIA_BIOS_NUM_TV_SPECIAL_REG];
    unsigned char   CRTCPAL2_8BPP[VIA_BIOS_MAX_NUM_TV_CRTC];
    unsigned char   CRTCPAL2_16BPP[VIA_BIOS_MAX_NUM_TV_CRTC];
    unsigned char   CRTCPAL2_32BPP[VIA_BIOS_MAX_NUM_TV_CRTC];
    unsigned int    PatchPAL2[VIA_BIOS_MAX_NUM_TV_PATCH];
    unsigned int    RGBPAL[VIA_BIOS_NUM_TV_OTHER];
    unsigned int    YCbCrPAL[VIA_BIOS_NUM_TV_OTHER];
    unsigned int    SDTV_RGBPAL[VIA_BIOS_NUM_TV_OTHER];
    unsigned int    SDTV_YCbCrPAL[VIA_BIOS_NUM_TV_OTHER];
} VIABIOSTV3TableRec, *VIABIOSTV3TablePtr;


#endif /* _VIA_BIOS_H */
