/* 
 * Via CLE266 Hardware Acceleration
 *
 * Copyright (C) 2004 Andre Kriehn
 * All Rights Reserved      
 */

#include <linux/wait.h>
#include <linux/fb.h>

#include "regs.h"
#include "via_fbobj.h"

int WaitIdleCLE266(struct fb_info *info)
{
    struct via_par * par = (struct via_par *)info->par;
    int loop = 0;

    while (!VIAGETREG(par->io_base_virt,VIA_REG_STATUS) & VIA_VR_QUEUE_BUSY) {
	if (++loop > MAXLOOP) break;
    }
//    while ((VIAGETREG(par->io_base_virt,VIA_REG_STATUS) & 
//	(VIA_CMD_RGTR_BUSY | VIA_VR_QUEUE_BUSY)) != VIA_CMD_RGTR_BUSY) {
//	if (++loop > MAXLOOP) break;
//    }
    return loop >= MAXLOOP;
}

void setGEMode(struct fb_info *info) {
    struct via_par * par = (struct via_par *)info->par;
    
    unsigned int dwGEMode = 0;

    switch (info->var.bits_per_pixel) {
    case 16:
	dwGEMode |= VIA_GEM_16bpp;
	break;
    case 32:
	dwGEMode |= VIA_GEM_32bpp;
	break;
    default:
	dwGEMode |= VIA_GEM_8bpp;
	break;
    }
    VIASETREG(par->io_base_virt, VIA_REG_GEMODE, dwGEMode);
}

int viafb_init_2d_engine(struct fb_info *info, const int enable)
{
    struct via_par * par = (struct via_par *)info->par;
    // Init 2D engine registers to reset 2D engine

    VIASETREG(par->io_base_virt, VIA_REG_GECMD,     0x0);
    VIASETREG(par->io_base_virt, VIA_REG_SRCPOS,    0x0);
    VIASETREG(par->io_base_virt, VIA_REG_DSTPOS,    0x0);
    VIASETREG(par->io_base_virt, VIA_REG_DIMENSION, 0x0);
    VIASETREG(par->io_base_virt, VIA_REG_PATADDR,   0x0);
    VIASETREG(par->io_base_virt, VIA_REG_FGCOLOR,   0x0);
    VIASETREG(par->io_base_virt, VIA_REG_BGCOLOR,   0x0);
    VIASETREG(par->io_base_virt, VIA_REG_CLIPTL,    0x0);
    VIASETREG(par->io_base_virt, VIA_REG_CLIPBR,    0x0);
    VIASETREG(par->io_base_virt, VIA_REG_OFFSET,    0x0);
    VIASETREG(par->io_base_virt, VIA_REG_KEYCONTROL,0x0);
    VIASETREG(par->io_base_virt, VIA_REG_SRCBASE,   0x0);
    VIASETREG(par->io_base_virt, VIA_REG_DSTBASE,   0x0);
    VIASETREG(par->io_base_virt, VIA_REG_PITCH,     0x0);
    VIASETREG(par->io_base_virt, VIA_REG_MONOPAT0,  0x0);
    VIASETREG(par->io_base_virt, VIA_REG_MONOPAT1,  0x0);

    // Init AGP and VQ registers

    VIASETREG(par->io_base_virt, VIA_REG_TRANSET,   0x00100000);
    VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x00000000);
    VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x00333004);
    VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x60000000);
    VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x61000000);
    VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x62000000);
    VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x63000000);
    VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x64000000);
    VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x7D000000);

    VIASETREG(par->io_base_virt, VIA_REG_TRANSET,   0xfe020000);
    VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x00000000);

    if (enable) { // Enable VQ
        VIASETREG(par->io_base_virt, VIA_REG_TRANSET,   0x00fe0000);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x080003fe);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x0a00027c);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x0b000260);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x0c000274);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x0d000264);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x0e000000);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x0f000020);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x1000027e);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x110002fe);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x200f0060);

        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x00000006);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x40008c0f);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x44000000);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x45080c04);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x46800408);

        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x52000000 |
            ((par->vq_start & 0xFF000000) >> 24) | ((par->vq_end & 0xFF000000) >> 16));
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x50000000 | (par->vq_start & 0xFFFFFF));
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x51000000 | (par->vq_end & 0xFFFFFF));
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x53000000 | (par->vq_size >> 3));
    }
    else { // Disable VQ

        VIASETREG(par->io_base_virt, VIA_REG_TRANSET,   0x00fe0000);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x00000004);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x40008c0f);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x44000000);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x45080c04);
        VIASETREG(par->io_base_virt, VIA_REG_TRANSPACE, 0x46800408);
    }

    setGEMode(info);
    /* Set BPP and Pitch */
    /* Set Src and Dst base address and pitch, pitch is qword */
    VIASETREG(par->io_base_virt,VIA_REG_SRCBASE, 0x0);
    VIASETREG(par->io_base_virt,VIA_REG_DSTBASE, 0x0);
    VIASETREG(par->io_base_virt,VIA_REG_PITCH, VIA_PITCH_ENABLE |
	      ((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) |
	      (((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) << 16));
    return 0;
}

int viafb_init_cursor(struct fb_info *info)
{
    struct via_par * par = (struct via_par *)info->par;
	
    VIASETREG(par->io_base_virt, VIA_REG_CURSOR_MODE, par->cur_start);
    memset((void*)par->mem_base_virt + par->cur_start,0xFF,par->cur_size);
    return 0;
}


/* ------------ Accelerated Functions --------------------- */
/**
 *      viafb_fillrect - REQUIRED function. Can use generic routines if 
 *		 	 non acclerated hardware and packed pixel based.
 *			 Draws a rectangle on the screen.		
 *
 *      @info: frame buffer structure that represents a single frame buffer
 *	@region: The structure representing the rectangular region we 
 *		 wish to draw to.
 *
 *	This drawing operation places/removes a retangle on the screen 
 *	depending on the rastering operation with the value of color which
 *	is in the current color depth format.
 */
void viafb_fillrect(struct fb_info *info, const struct fb_fillrect *region)
{
/*	Meaning of struct fb_fillrect
 *
 *	@dx: The x and y corrdinates of the upper left hand corner of the 
 *	@dy: area we want to draw to. 
 *	@width: How wide the rectangle is we want to draw.
 *	@height: How tall the rectangle is we want to draw.
 *	@color:	The color to fill in the rectangle with. 
 *	@rop: The raster operation. We can draw the rectangle with a COPY
 *	      of XOR which provides erasing effect. 
 */
    unsigned int col;
    struct via_par * par = (struct via_par *)info->par;
  
    debug("Fillrect dx:%d dy:%d width:%d height:%d col:%d",
	region->dx,region->dy,region->width,region->height,region->color);
  
    if (!(region->width && region->height)) return;

    col = 0;
    switch (info->var.bits_per_pixel) {
    case  8: col=region->color; 
	     break;
    case 15:
    case 16:
    case 24:
    case 32: col=((u32 *) (info->pseudo_palette))[region->color];
	     break;
    }
    
    WaitIdleCLE266(info);
       
    VIASETREG(par->io_base_virt, VIA_REG_SRCBASE, 0x0);
    VIASETREG(par->io_base_virt, VIA_REG_DSTBASE, 0x0);
    VIASETREG(par->io_base_virt, VIA_REG_PITCH, VIA_PITCH_ENABLE |
	      ((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) |
	      (((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) << 16));

    setGEMode(info);
    VIASETREG(par->io_base_virt, VIA_REG_DSTPOS,    ((region->dy) << 16) | (region->dx));
    VIASETREG(par->io_base_virt, VIA_REG_DIMENSION, ((region->height - 1) << 16) | (region->width - 1));

    if (region->rop == ROP_COPY) {
    
	VIASETREG(par->io_base_virt, VIA_REG_FGCOLOR, col);
	VIASETREG(par->io_base_virt, VIA_REG_GECMD, VIA_GEC_BLT | 
	    VIA_GEC_FIXCOLOR_PAT | VIA_ROP_P); // | VIA_GEC_CLIP_ENABLE
    } else {
	VIASETREG(par->io_base_virt, VIA_REG_GECMD, 
	    VIA_GEC_BLT | VIA_ROP_P); // | VIA_GEC_CLIP_ENABLE
    }
    return;
}

/**
 *      viafb_copyarea - REQUIRED function. Can use generic routines if
 *                       non acclerated hardware and packed pixel based.
 *                       Copies one area of the screen to another area.
 *
 *      @info: frame buffer structure that represents a single frame buffer
 *      @area: Structure providing the data to copy the framebuffer contents
 *	       from one region to another.
 *
 *      This drawing operation copies a rectangular area from one area of the
 *	screen to another area.
 */
void viafb_copyarea(struct fb_info *info, const struct fb_copyarea *area) 
{
/*
 *      @dx: The x and y coordinates of the upper left hand corner of the
 *	@dy: destination area on the screen.
 *      @width: How wide the rectangle is we want to copy.
 *      @height: How tall the rectangle is we want to copy.
 *      @sx: The x and y coordinates of the upper left hand corner of the
 *      @sy: source area on the screen.
 */
    
    unsigned int sx = area->sx;
    unsigned int sy = area->sy;
    unsigned int dx = area->dx;
    unsigned int dy = area->dy;
    unsigned int h = area->height;
    unsigned int w = area->width;
    
    unsigned int cmd = VIA_GEC_BLT | VIA_ROP_S; // | VIA_GEC_CLIP_ENABLE
    
    struct via_par * par = (struct via_par *)info->par;
    
    debug("Copyarea sx:%d sy:%d dx:%d dy:%d width:%d height:%d\n",
	area->sx,area->sy,area->dx,area->dy,area->width,area->height);
    
    if (!(w && h))
	return;
    
    if (sx < dx) {
	cmd |= VIA_GEC_DECX;
	sx += (w - 1);
	dx += (w - 1);
    }

    if (sy < dy) {
	cmd |= VIA_GEC_DECY;
	sy += (h - 1);
	dy += (h - 1);
    }

    /* command will be lost if we don't wait engine idle */
    WaitIdleCLE266(info);

    /* Set Src and Dst base address and pitch, pitch is qword */
    VIASETREG(par->io_base_virt,VIA_REG_SRCBASE, 0x0);
    VIASETREG(par->io_base_virt,VIA_REG_DSTBASE, 0x0);
    VIASETREG(par->io_base_virt,VIA_REG_PITCH, VIA_PITCH_ENABLE |
	      ((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) |
	      (((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) << 16));
    
    setGEMode(info);
    VIASETREG(par->io_base_virt,VIA_REG_SRCPOS, ((sy << 16) | sx));
    VIASETREG(par->io_base_virt,VIA_REG_DSTPOS, ((dy << 16) | dx));
    VIASETREG(par->io_base_virt,VIA_REG_DIMENSION, (((h - 1) << 16) | (w - 1)));
    VIASETREG(par->io_base_virt,VIA_REG_GECMD, cmd);
    return;
}


void viafb_monoblit(struct fb_info *info, const struct fb_image *image) 
{
    struct via_par * par = (struct via_par *)info->par;
    
    unsigned int dx= image->dx;
    unsigned int dy= image->dy;
    unsigned int h = image->height;
    unsigned int w = image->width;
    u8* data=(u8*)image->data;
    
    unsigned int cmd = VIA_GEC_BLT | VIA_GEC_SRC_SYS | VIA_ROP_S | 
		       VIA_GEC_SRC_MONO | VIA_GEC_MONO_PACK;
//    printk("Img cmd:%X\n",cmd);
    /* command will be lost if we don't wait engine idle */
    WaitIdleCLE266(info);
    
//    printk("Img depth:%d x:%d y:%d h:%d w:%d\n",image->depth,dx,dy,h,w);
    /* Set Src and Dst base address and pitch, pitch is qword */
    VIASETREG(par->io_base_virt,VIA_REG_SRCBASE, 0x0);//(unsigned int)image->data
    VIASETREG(par->io_base_virt,VIA_REG_DSTBASE, 0x0);
    VIASETREG(par->io_base_virt,VIA_REG_PITCH, VIA_PITCH_ENABLE |
	      ((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) |
	      (((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) << 16));
  
    setGEMode(info);
    VIASETREG(par->io_base_virt,VIA_REG_KEYCONTROL, 0x0);

    VIASETREG(par->io_base_virt,VIA_REG_SRCPOS, 0);
    VIASETREG(par->io_base_virt,VIA_REG_DSTPOS, ((dy << 16) | dx));
    VIASETREG(par->io_base_virt,VIA_REG_DIMENSION, (((h - 1) << 16) | (w - 1)));
    switch (info->var.bits_per_pixel) {
	case  8: VIASETREG(par->io_base_virt, VIA_REG_BGCOLOR, image->bg_color);
		 VIASETREG(par->io_base_virt, VIA_REG_FGCOLOR, image->fg_color);
	         break;
        case 15:
	case 16: 
        case 24:
	case 32: VIASETREG(par->io_base_virt, VIA_REG_BGCOLOR, ((u32 *) (info->pseudo_palette))[image->bg_color]);
		 VIASETREG(par->io_base_virt, VIA_REG_FGCOLOR, ((u32 *) (info->pseudo_palette))[image->fg_color]);
	         break;
    }
    
    VIASETREG(par->io_base_virt,VIA_REG_GECMD, cmd);
  
    memcpy((u8*)(par->io_base_virt + VIA_MMIO_BLTBASE), 
	data, (w * h)>>3 );

}

void viafb_colorblit(struct fb_info *info, const struct fb_image *image) 
{
    struct via_par * par = (struct via_par *)info->par;        
    
    unsigned int dx= image->dx;
    unsigned int dy= image->dy;
    unsigned int h = image->height;
    unsigned int w = image->width;
    u8* data=(u8*)image->data;
    
    unsigned int cmd = VIA_GEC_BLT | VIA_GEC_SRC_SYS | VIA_ROP_S;
//    printk("Img cmd:%X\n",cmd);
    /* command will be lost if we don't wait engine idle */
    WaitIdleCLE266(info);
    
//    printk("Img depth:%d x:%d y:%d h:%d w:%d\n",image->depth,dx,dy,h,w);
    /* Set Src and Dst base address and pitch, pitch is qword */
    VIASETREG(par->io_base_virt,VIA_REG_SRCBASE, 0x0);//(unsigned int)image->data
    VIASETREG(par->io_base_virt,VIA_REG_DSTBASE, 0x0);
    VIASETREG(par->io_base_virt,VIA_REG_PITCH, VIA_PITCH_ENABLE |
	      ((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) |
	      (((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) << 16));
  
    setGEMode(info);
    VIASETREG(par->io_base_virt,VIA_REG_KEYCONTROL, 0x0);

    VIASETREG(par->io_base_virt,VIA_REG_SRCPOS, 0);
    VIASETREG(par->io_base_virt,VIA_REG_DSTPOS, ((dy << 16) | dx));
    VIASETREG(par->io_base_virt,VIA_REG_DIMENSION, (((h - 1) << 16) | (w - 1)));
    switch (info->var.bits_per_pixel) {
	case  8: VIASETREG(par->io_base_virt, VIA_REG_BGCOLOR, image->bg_color);
		 VIASETREG(par->io_base_virt, VIA_REG_FGCOLOR, image->fg_color);
	         break;
        case 15:
	case 16: 
        case 24:
	case 32: VIASETREG(par->io_base_virt, VIA_REG_BGCOLOR, ((u32 *) (info->pseudo_palette))[image->bg_color]);
		 VIASETREG(par->io_base_virt, VIA_REG_FGCOLOR, ((u32 *) (info->pseudo_palette))[image->fg_color]);
	         break;
    }
    
    VIASETREG(par->io_base_virt,VIA_REG_GECMD, cmd);
  
    memcpy((u8*)(par->io_base_virt + VIA_MMIO_BLTBASE), 
	data, (w * image->depth * h)>>3);

}
/**
 *      viafb_imageblit - REQUIRED function. Can use generic routines if
 *                        non acclerated hardware and packed pixel based.
 *                        Copies a image from system memory to the screen. 
 *
 *      @info: frame buffer structure that represents a single frame buffer
 *	@image:	structure defining the image.
 *
 *      This drawing operation draws a image on the screen. It can be a 
 *	mono image (needed for font handling) or a color image (needed for
 *	tux). 
 */
void viafb_imageblit(struct fb_info *info, const struct fb_image *image) 
{
/*
 *      @dx: The x and y coordinates of the upper left hand corner of the
 *	@dy: destination area to place the image on the screen.
 *      @width: How wide the image is we want to copy.
 *      @height: How tall the image is we want to copy.
 *      @fg_color: For mono bitmap images this is color data for     
 *      @bg_color: the foreground and background of the image to
 *		   write directly to the frmaebuffer.
 *	@depth:	How many bits represent a single pixel for this image.
 *	@data: The actual data used to construct the image on the display.
 *	@cmap: The colormap used for color images.   
 */
//    struct via_par * par = (struct via_par *)info->par;
    
    switch (image->depth) {
    case 1: viafb_monoblit(info,image);
	    break;
    case 15:
    case 16:
    case 24:
    case 32:printk("viafb: colorblit depth %d\n",image->depth); 
	    //viafb_colorblit(info,image);
	    cfb_imageblit(info, image);
	    break;
    default: cfb_imageblit(info, image);
    }
}

void viafb_cursorimg(struct fb_info *info, const struct fb_image *image) 
{
    struct via_par * par = (struct via_par *)info->par;
    
    unsigned int h = image->height;
    unsigned int w = image->width;
    u8* data=(u8*)image->data;
    
    unsigned int cmd = VIA_GEC_BLT | VIA_GEC_SRC_SYS | VIA_ROP_S | 
		       VIA_GEC_SRC_MONO | VIA_GEC_MONO_PACK;

    /* command will be lost if we don't wait engine idle */
    WaitIdleCLE266(info);
    
    /* Set Src and Dst base address and pitch, pitch is qword */
    VIASETREG(par->io_base_virt,VIA_REG_SRCBASE, 0x0);
    VIASETREG(par->io_base_virt,VIA_REG_DSTBASE, par->cur_start);
    VIASETREG(par->io_base_virt,VIA_REG_PITCH, VIA_PITCH_ENABLE |
	      ((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) |
	      (((info->var.xres * info->var.bits_per_pixel >> 3) >> 3) << 16));
  
    setGEMode(info);
    VIASETREG(par->io_base_virt,VIA_REG_KEYCONTROL, 0x0);

    VIASETREG(par->io_base_virt,VIA_REG_SRCPOS, 0);
    VIASETREG(par->io_base_virt,VIA_REG_DSTPOS, 0);
    VIASETREG(par->io_base_virt,VIA_REG_DIMENSION, (((h - 1) << 16) | (w - 1)));
    switch (info->var.bits_per_pixel) {
	case  8: VIASETREG(par->io_base_virt, VIA_REG_BGCOLOR, image->bg_color);
		 VIASETREG(par->io_base_virt, VIA_REG_FGCOLOR, image->fg_color);
	         break;
        case 15:
	case 16: 
        case 24:
	case 32: VIASETREG(par->io_base_virt, VIA_REG_BGCOLOR, ((u32 *) (info->pseudo_palette))[image->bg_color]);
		 VIASETREG(par->io_base_virt, VIA_REG_FGCOLOR, ((u32 *) (info->pseudo_palette))[image->fg_color]);
	         break;
    }
    VIASETREG(par->io_base_virt,VIA_REG_GECMD, cmd);
    memcpy((u8*)(par->io_base_virt + VIA_MMIO_BLTBASE), 
	data, (w * h)>>3 );
}

/**
 *	viafb_cursor - 	REQUIRED function. If your hardware lacks support
 *			for a cursor you can use the default cursor whose
 *			function is called soft_cursor. It will always 
 *			work since it uses viafb_imageblit function which 
 *			is required. 	  	 
 *
 *      @info: frame buffer structure that represents a single frame buffer
 *	@cursor: structure defining the cursor to draw.
 *
 *      This operation is used to set or alter the properities of the
 *	cursor.
 *
 *	Returns negative errno on error, or zero on success.
 */
int viafb_cursor(struct fb_info *info, struct fb_cursor *cursor)
{
    u32 oldMode;
    u8 *data;
    struct via_par * p = (struct via_par *)info->par;
    
    int size;
    int i,y;
    unsigned long mem;
    unsigned long count;
/*
 *      @set: 	Which fields we are altering in struct fb_cursor 
 *	@enable: Disable or enable the cursor 
 *      @rop: 	The bit operation we want to do. 
 *      @mask:  This is the cursor mask bitmap. 
 *      @dest:  A image of the area we are going to display the cursor.
 *		Used internally by the driver.	 
 *      @hot:	The hot spot. 
 *	@image:	The actual data for the cursor image.
 */
    
    /* Hide cursor before set cursor position in order to avoid ghost cursor
     * image when directly set cursor position. It should be a HW bug but
     * we can use patch by SW. */
    oldMode = VIAGETREG(p->io_base_virt, VIA_REG_CURSOR_MODE);

    /* Turn cursor off. */
    VIASETREG(p->io_base_virt, VIA_REG_CURSOR_MODE, oldMode & 0xFFFFFFFE);
    
    if (cursor->set & FB_CUR_SETSIZE) {
//	info->cursor.image.height = cursor->image.height;
//	info->cursor.image.width = cursor->image.width;
    }
 
    if (cursor->set & FB_CUR_SETCMAP) {
//	info->cursor.image.fg_color = cursor->image.fg_color;
//	info->cursor.image.bg_color = cursor->image.bg_color;

	switch (info->var.bits_per_pixel) {
	case  8: VIASETREG(p->io_base_virt, VIA_REG_CURSOR_BG, cursor->image.bg_color);
		 VIASETREG(p->io_base_virt, VIA_REG_CURSOR_FG, cursor->image.fg_color);
	         break;
        case 15:
	case 16: 
        case 24:
	case 32: VIASETREG(p->io_base_virt, VIA_REG_CURSOR_BG, ((u32 *) (info->pseudo_palette))[cursor->image.bg_color]);
		 VIASETREG(p->io_base_virt, VIA_REG_CURSOR_FG, ((u32 *) (info->pseudo_palette))[cursor->image.fg_color]);
	         break;
	}	
//	info->cursor.image.depth = cursor->image.depth;
    }
    
    /*VIASETREG(p->io_base_virt, VIA_REG_CURSOR_BG, ((u16 *) (info->pseudo_palette))[cursor->image.bg_color]);
		 VIASETREG(p->io_base_virt, VIA_REG_CURSOR_FG, ((u16 *) (info->pseudo_palette))[cursor->image.fg_color]);
		 break;
    */
    if (cursor->set & FB_CUR_SETPOS) {
	u32 tmp;

//	info->cursor.image.dx = cursor->image.dx;
//	info->cursor.image.dy = cursor->image.dy;
		
	tmp = (cursor->image.dx - info->var.xoffset) << 16;
	tmp |= (cursor->image.dy - info->var.yoffset);

/*	if (((cursor->image.dx - info->var.xoffset) < 0) ||
	    ((cursor->image.dy - info->var.yoffset) < 0)) {
	    printk("fb: x:%d y:%d ox:%d oy:%d \n",cursor->image.dx,cursor->image.dy,
		info->var.xoffset,info->var.yoffset);
	}
*/	
	VIASETREG(p->io_base_virt, VIA_REG_CURSOR_ORG, 0);//((xoff << 16) | (yoff & 0x003f))
        VIASETREG(p->io_base_virt, VIA_REG_CURSOR_POS, tmp);//((x << 16) | (y & 0x07ff))
    } 
    
//    info->cursor.image.data = cursor->image.data;

    if (info->state != FBINFO_STATE_RUNNING)
	return 0;

    size = ((cursor->image.width + 7) >> 3) * 
    	cursor->image.height;
	
//    printk("Cursor width:%d height:%d depth:%d size:%d\n",info->cursor.image.width,
//	info->cursor.image.height,info->cursor.image.depth,size);
    data = kmalloc(64 + sizeof(struct fb_cursor), GFP_ATOMIC);
    if (!data)
	return -ENOMEM;

    if (cursor->enable) {
	switch (cursor->rop) {
	    case ROP_XOR:
		for (i = 0; i < size; i++)
		    data[i] = cursor->image.data[i] ^ cursor->mask[i]; 
		break;
	    case ROP_COPY:
	    default:
		for (i = 0; i < size; i++)
		    data[i] = cursor->image.data[i] & cursor->mask[i]; 
    		break;
	}
	mem = p->mem_base_virt + p->cur_start;
	count=0;

	for (y=0; y< size; y++) {//
	      *(volatile u32*)(mem + count)  = data[y];
	      *(volatile u32*)(mem + count+4)= 0xFFFFFFFF- data[y];
	    count+=8;
	}
	for (y=0; y< size; y++) {//
	    *(volatile u32*)(mem + count)  = 0x00000000;
	    *(volatile u32*)(mem + count+4)= 0xFFFFFFFF;
	    count+=8;
	}
//	viafb_cursorimg(info, (struct fb_image *)&(cursor->image));
        /* Turn on Hardware Cursor */
        VIASETREG(p->io_base_virt, VIA_REG_CURSOR_MODE, oldMode | 0x3);
    }   
    if (data) kfree(data);
 
    return 0;	
}

