/*
 * Copyright (C) 2003 VIA Technologies, Inc. All Rights Reserved.
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU *General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) any later 
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Port to kernel 2.6 by Andre Kriehn <AndreKriehn@gmx.de> 
 *
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/wait.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <asm/mtrr.h>
#include <asm/uaccess.h>

#include "via_fbobj.h"
#include "regs.h"
#include "via_i2c.h"
#include "via_bios.h"
#include "via_mode.h"
#include "via_tv3.h"
#include "via_vt1622a.h"

#ifndef FB_ACCEL_VIA_UNICHROME
#define FB_ACCEL_VIA_UNICHROME 77
#endif

#ifndef FBIO_WAITFORVSYNC
#define FBIO_WAITFORVSYNC   _IOW('F', 0x20, u_int32_t)
#endif
#ifndef FBIO_FLIPONVSYNC
#define FBIO_FLIPONVSYNC    _IOWR('F', 0x21, struct fb_flip)
#endif

static struct fb_fix_screeninfo viafb_fix = {
    .id =       "UNICHROME", 
    .type =     FB_TYPE_PACKED_PIXELS,
    .visual =   FB_VISUAL_PSEUDOCOLOR,
    .xpanstep = 0,
    .ypanstep = 1,
    .ywrapstep = 0, 
    .accel =    FB_ACCEL_VIA_UNICHROME,
};

static struct fb_info* info;
static struct via_par* current_par;

#ifndef VIA_MMIO
#define VIA_MMIO 0
#endif

#if VIA_MMIO
#define t_outb(val, reg)    writeb(val, current_par.io_base_virt + reg)
#define t_inb(reg)          readb(current_par.io_base_virt + reg)
#else
#define t_outb(val,reg)     outb(val,reg)
#define t_inb(reg)          inb(reg)
#endif

#define VERSION     "1.0"

int VIACheckTVExist    (struct fb_info *info);
void VIAPreSetTV3Mode  (struct fb_info *info,int xres,int yres,int bpp);
void VIAPostSetTV3Mode (struct fb_info *info,int xres,int yres,int bpp);
static int viafb_pan_display  (struct fb_var_screeninfo *var, struct fb_info *info);
static int viafb_sync         (struct fb_info *info);
#ifndef MODULE
static int viafb_setup        (char *options);
#endif
static int viafb_wait_for_sync(struct fb_info *info, int field_option);

static struct { unsigned char red,green,blue,transp; } palette[256];

static u32 pseudo_pal[16];

/* video mode */
static char * mode    = "720x576";
static int bpp        = 32;
static int refresh    = 0xFF;
static int resMode    = VIA_RES_720X576;

static int noaccel    = 0;
static int memsize    = 0;
static int TVon       = 0;
static int TVtype     = TVTYPE_PAL; // 0:none  1:NTSC  2:PAL
static int TVoverscan = VIA_TVNORMAL; // 0:VIA_TVNORMAL  1:VIA_TVOVER
static int TVEncoder  = VIA_NONETV;

extern void dmp_vq(struct fb_info *info);
extern int viafb_init_2d_engine(struct fb_info *info, const int enable);
extern int viafb_init_cursor(struct fb_info *info);
extern void viafb_fillrect(struct fb_info *info, const struct fb_fillrect *region);
extern void viafb_copyarea(struct fb_info *info, const struct fb_copyarea *area);
extern void viafb_imageblit(struct fb_info *info, const struct fb_image *image);
extern int viafb_cursor(struct fb_info *info, struct fb_cursor *cursor);
extern int WaitIdleCLE266(struct fb_info *info);

#ifndef wait_event_interruptible_timeout
#define __wait_event_interruptible_timeout(wq, condition, ret)      \
do {                                                                \
    wait_queue_t __wait;                                            \
    init_waitqueue_entry(&__wait, current);                         \
                                                                    \
    add_wait_queue(&wq, &__wait);                                   \
    for (;;) {                                                      \
        set_current_state(TASK_INTERRUPTIBLE);                      \
        if (condition)                                              \
            break;                                                  \
        if (!signal_pending(current)) {                             \
            ret = schedule_timeout(ret);                            \
            if (!ret)                                               \
                break;                                              \
            continue;                                               \
        }                                                           \
        ret = -ERESTARTSYS;                                         \
        break;                                                      \
    }                                                               \
    current->state = TASK_RUNNING;                                  \
    remove_wait_queue(&wq, &__wait);                                \
} while (0)

#define wait_event_interruptible_timeout(wq, condition, timeout)    \
({                                                                  \
    long __ret = timeout;                                           \
    if (!(condition))                                               \
        __wait_event_interruptible_timeout(wq, condition, __ret);   \
    __ret;                                                          \
})
#endif

static inline unsigned char read3X4(int reg)
{
    outb(reg, 0x3D4);
    return inb(0x3D5);
}

static inline void write3X4(int reg, unsigned char val)
{
    outb(reg, 0x3D4);
    outb(val, 0x3D5);
}

static inline unsigned char read3C4(int reg)
{
    t_outb(reg, 0x3C4);
    return t_inb(0x3C5);
}

static inline void write3C4(int reg, unsigned char val)
{
    t_outb(reg, 0x3C4);
    t_outb(val, 0x3C5);
}

static inline unsigned char read3CE(int reg)
{
    t_outb(reg, 0x3CE);
    return t_inb(0x3CF);
}

static inline void write3CE(int reg, unsigned char val)
{
    t_outb(reg, 0x3CE);
    t_outb(val, 0x3CF);
}

static inline void writeAttr(int reg, unsigned char val)
{
    t_inb(0x3DA);
    t_outb(reg, 0x3C0);
    t_outb(val, 0x3C0);
}

static inline unsigned char readAttr(int reg)
{
    t_inb(0x3DA);
    t_outb(reg, 0x3C0);
    return t_inb(0x3C1);
}

static void via_enable_mmio(void)
{
    outb(inb(0x3c3) | 0x01, 0x3c3);
    outb(inb(0x3cc) | 0x01, 0x3c2);
    /* unlock extended register */
    outb(0x10, 0x3c4);
    outb(0x01, 0x3c5);
    /* Enable MMIO */
//    write3C4(0x1a, read3C4(0x1a) | 0x68);
    outb(0x1a, 0x3c4);
    outb(inb(0x3c5) | 0x68, 0x3c5);
}

static void via_disable_mmio(void)
{
//    outb(0x1a, 0x3c4);
//    outb(inb(0x3c5) & 0x97, 0x3c5);
}

static void via_set_regs(void) 
{
    outb(0x1a, 0x3c4);
    outb(inb(0x3c5) & 0xBF, 0x3c5);
    outb(inb(0x3c5) | 0x04, 0x3c5);
}

static unsigned int get_memsize(void)
{
    unsigned int m;

    /* If memory size provided by user */
    if (memsize)
        m = memsize * Mb;
    else {
        write3C4(0x10,1);
        outb(0x34, 0x3c4);  // for register CR34 (Memory Size)
        m = inb(0x3c5);
        if (m > 16 && m <= 128)
            m = (inb(0x3c5)+1)<<19;
        else {
            outb(0x39, 0x3c4);
            m = inb(0x3c5);
            if (m > 16 && m <= 128)
                m = (m + 1) << 19;
            else if (m > 0 && m < 31)
                m = m << 22;
            else
                m = 16 * Mb;
        }
    }
    output("framebuffer size = %d Mb\n", m/Mb);
    return m;
}

static int setmode(struct fb_info *info,int x,int y,int b)
{
    const struct via_par *par = (const struct via_par *) info->par;
    
    int selectmode = -1;
    int j, k = 0, m, n;
    int port, offset, mask, data;
    int countWidthByQWord, offsetWidthByQWord;
    unsigned char cr13, cr35, sr1c, sr1d;

    if (b==24)   b = 32;

    for (j=0; j<NumModes ; j++){
        if ( x==Modes[j].Width && y==Modes[j].Height && b==Modes[j].Bpp ) {
            selectmode=j;
            break;
        }
    }
    if (selectmode < 0) {
        debug("No mode select,No support mode : %dx%d\n",x,y);
        return -1;
    }
    debug("setmode fount mode:%d\n",selectmode);
    if (par->TVon && par->TVtype!=TVTYPE_NONE) {
        debug("setmode TVon && TVtype!=TVTYPE_NONE\n");
        switch (par->TVEncoder) {
            case VIA_TV3:
            case VIA_VT1622A:
                VIAPreSetTV3Mode(info,x,y,b);
                break;
        }
    }
    /* Turn off Screen */
    write3X4(0x17, read3X4(0x17) & 0x7f);
  
    /* Set Sequences regs */
    for (j = 1; j < 5; j++) {
        write3C4(j, Modes[selectmode].stdVgaTable.SR[j]);
    }
    /* Set Misc reg */
    t_outb(Modes[selectmode].stdVgaTable.misc,0x3c2);
    /* Set CRTC regs */
    for (j = 0; j < 25; j++) {
        write3X4(j,Modes[selectmode].stdVgaTable.CR[j]);
    }
    /* Set attribute regs */
    for (j = 0; j < 20; j++) {
        writeAttr(j,Modes[selectmode].stdVgaTable.AR[j]);
    }
    for (j = 0; j < 9; j++) {
        write3CE(j,Modes[selectmode].stdVgaTable.GR[j]);
    }

    /* Unlock Extended regs */
    write3C4(0x10, 0x01);
    
    if (Modes[selectmode].Mode <= 0x13) {
        /* Set Standard Mode-Spec. Extend Regs */
        for (j = 0; j < stdModeExtTable.numEntry; j++) {
            port = stdModeExtTable.port[j];
            offset = stdModeExtTable.offset[j];
            mask = stdModeExtTable.mask[j];
            data = stdModeExtTable.data[j] & mask;
            t_outb(offset, 0x300+port);
            t_outb(data, 0x301+port);
        }
    }
    else {
        /* Set Extended Mode-Spec. Extend Regs */
        for (j = 0; j < Modes[selectmode].extModeExtTable.numEntry; j++) {
            port = Modes[selectmode].extModeExtTable.port[j];
            offset = Modes[selectmode].extModeExtTable.offset[j];
            mask = Modes[selectmode].extModeExtTable.mask[j];
            data = Modes[selectmode].extModeExtTable.data[j] & mask;
            t_outb(offset, 0x300+port);
            t_outb(data, 0x301+port);
        }
    }

    j = resMode;
    if ((j != VIA_RES_INVALID) && (par->refresh != 0xFF) && !par->TVon) {
        switch (par->refresh) {
            case 75:
                k = 0;
                break;
            case 85:
                k = 1;
                break;
            case 100:
                k = 2;
                break;
            case 120:
                k = 3;
                break;
        }
     
        write3X4(0x17, read3X4(0x17) & 0x7F);

        write3C4(0x46, (refreshTable[j][k].VClk >> 8));
        write3C4(0x47, (refreshTable[j][k].VClk & 0xFF));

        write3X4(0x17, read3X4(0x17) | 0x80);

        write3C4(0x40, read3C4(0x40) | 0x02);
        write3C4(0x40, read3C4(0x40) & 0xFD);

        m = 0;
        write3X4(0x0, refreshTable[j][k].CR[m++]);

        for (n = 2; n <= 7; n++)
            write3X4(n, refreshTable[j][k].CR[m++]);

        for (n = 16; n <= 17; n++)
            write3X4(n, refreshTable[j][k].CR[m++]);

        for (n = 21; n <= 22; n++)
            write3X4(n, refreshTable[j][k].CR[m++]);

        /* Use external clock */
        t_outb(t_inb(0x3cc) | 0x0C, 0x3c2);

    } else if (Modes[selectmode].VClk != 0) {
        write3X4(0x17, read3X4(0x17) & 0x7F);

        write3C4(0x46, (Modes[selectmode].VClk >> 8));
        write3C4(0x47, (Modes[selectmode].VClk & 0xFF));

        write3X4(0x17, read3X4(0x17) | 0x80);

        write3C4(0x40, read3C4(0x40) | 0x02);
        write3C4(0x40, read3C4(0x40) & 0xFD);

        /* Use external clock */
        t_outb(t_inb(0x3cc) | 0x0C, 0x3c2);
    }

    /* Set Quadword offset and counter */
    /* Fix bandwidth problem when using virtual desktop */
    countWidthByQWord = (x * (b >> 3)) >> 3;
    offsetWidthByQWord = (x * (b >> 3)) >> 3;

    cr35 = read3X4(0x35) & 0x1f;

    cr13 = offsetWidthByQWord & 0xFF;
    cr35 |= (offsetWidthByQWord & 0x700) >> 3;        /* bit 7:5: offset 10:8 */

    sr1d = read3C4(0x1d);

    /* Enable Refresh to avoid data lose when enter screen saver */
    /* Refresh disable & 128-bit alignment */
    sr1d = (sr1d & 0xfc) | (countWidthByQWord >> 9);
    sr1c = ((countWidthByQWord >> 1) + 1) & 0xff;
    
    write3X4(0x13, cr13);
    write3X4(0x35, cr35);
    write3C4(0x1c, sr1c);
    write3C4(0x1d, sr1d);

    write3X4(0x32, read3X4(0x32) | 1 );

    /* Open Screen */
    t_inb(0x3da);
    t_outb(0x20, 0x3c0);

    if (par->TVon && par->TVtype!=TVTYPE_NONE) {
        switch (par->TVEncoder) {
            case VIA_TV3:
            case VIA_VT1622A:
                VIAPostSetTV3Mode(info,x,y,b);
                break;
        }
    }

    /* must now re-enable vsync interrupt */
    outb(0x11, 0x3d4);
    outb(inb(0x3d5) | 0x30, 0x3d5);

    via_set_regs();
    return 1;
}

static int checkmode(int x,int y,int b)
{
    int selectmode = -1;
    int j;

    debug("checkmode\n");
    debug("check x:%d  y:%d  bpp:%d\n",x,y,b);
    
    if (b == 24)
        b = 32;

    for (j=0; j<NumModes ; j++) {
        if ((x==Modes[j].Width) && (y==Modes[j].Height) && (b==Modes[j].Bpp)) {
            selectmode=j;
            break;
        }
    }
    debug("selmode :%d\n",selectmode);
    if (selectmode < 0) {
        debug("No mode select,No support mode : %dx%d\n",x,y);
        return -EINVAL;
    }   
    return 0;
}

static int viafb_open(struct fb_info *info, int user)
{
    return 0;
}

static int viafb_release(struct fb_info *info, int user)
{
    return 0;
}

static int viafb_check_var(struct fb_var_screeninfo *var, struct fb_info *info)
{
    const struct via_par *par = (const struct via_par *) info->par;
    int result;
    
    debug("Check_var x:%d y:%d bpp:%d\n",var->xres,var->yres,var->bits_per_pixel);
    result = checkmode(var->xres,var->yres,var->bits_per_pixel);
    if (result < 0)
        return -EINVAL;
    if ((var->xres_virtual * (var->bits_per_pixel >> 3)*var->yres_virtual) > par->video_mem_size) {
        debug("check_var :%d  mem_size: %d\n",
            (var->xres_virtual * (var->bits_per_pixel >> 3) * var->yres_virtual),
            par->video_mem_size); 
        return -EINVAL;
    }
    return 0;
}

static int viafb_set_par(struct fb_info *info)
{
    struct via_par *par = (struct via_par *) info->par;
    int res=0;
    int size=0;
//    int vfront, vback, vsync, vres, hres;

    debug("Set_Par x:%d y:%d bpp:%d vxres:%d vyres:%d\n",info->var.xres,info->var.yres,
        info->var.bits_per_pixel,info->var.xres_virtual,info->var.yres_virtual);
       
    par->TVoverscan = (info->var.sync & FB_SYNC_BROADCAST) ? VIA_TVOVER : VIA_TVNORMAL;
    
    res = setmode(info,info->var.xres,info->var.yres,info->var.bits_per_pixel);
    if (res>0) {
        if (info->var.xres_virtual < info->var.xres) 
            info->var.xres_virtual = info->var.xres;
        if (info->var.yres_virtual < info->var.yres) 
            info->var.yres_virtual = info->var.yres;
        info->fix.line_length=info->var.xres * (info->var.bits_per_pixel >> 3);
        info->var.activate = FB_ACTIVATE_NOW;

#if 0
        info->var.pixclock=10000000/info->var.xres*1000/info->var.yres;
        info->var.left_margin=(info->var.xres>>3)&0xf8;
        info->var.right_margin=32;
        info->var.upper_margin=16;
        info->var.lower_margin=4;
        info->var.hsync_len=info->var.left_margin;
        info->var.vsync_len=4;
        info->var.accel_flags = 0;

        // convert from picoseconds to MHz 
        par->vclk = 1000000/info->var.pixclock;

        if (info->var.bits_per_pixel == 32)
            par->vclk *=2;
      
        vres  = info->var.yres;
        vfront = var->upper_margin;
        vback =   var->lower_margin;
        vsync =   var->vsync_len;

        // Compute horizontal and vertical VGA CRTC timing values 
        if (info->var.vmode & FB_VMODE_INTERLACED) {
            vres /= 2;
            vfront /=2;
            vback /=2;
            vsync /=2;
        }

        if (info->var.vmode & FB_VMODE_DOUBLE) {
            vres *= 2;
            vfront *=2;
            vback *=2;
            vsync *=2;
        }

        info->var.htotal = (info->var.xres + info->var.left_margin + info->var.right_margin + info->var.hsync_len)/8 - 10;
        info->var.hdispend = info->var.xres/8 - 1;
        info->var.hsyncstart = (info->var.xres + info->var.right_margin)/8;
        info->var.hsyncend = info->var.hsync_len/8;
        info->var.hblankstart = info->var.hdispend + 1;
        info->var.hblankend = info->var.htotal + 5;

        info->var.vtotal = vres + vfront + vback + vsync - 2;
        info->var.vdispend = vres - 1;
        info->var.vsyncstart = vres + vback;
        info->var.vsyncend = vsync;
        info->var.vblankstart = vres;
        info->var.vblankend = info->var.vtotal + 2;
#endif

        switch (info->var.bits_per_pixel) {
            case 8:
                info->var.red.offset = 0;
                info->var.green.offset = 0;
                info->var.blue.offset = 0;
                info->var.blue.offset = 0;
                info->var.red.length = 6;
                info->var.green.length = 6;
                info->var.blue.length = 6;
                info->var.transp.length = 0;
                info->fix.visual = FB_VISUAL_PSEUDOCOLOR;
                break;
            case 16:
                info->var.red.offset = 11;
                info->var.green.offset = 5;
                info->var.blue.offset = 0;
                info->var.transp.offset = 0;
                info->var.red.length = 5;
                info->var.green.length = 6;
                info->var.blue.length = 5;
                info->var.transp.length = 0;
                info->fix.visual = FB_VISUAL_TRUECOLOR;
                break;
            case 24:
            case 32:
                info->var.red.offset = 16;
                info->var.green.offset = 8;
                info->var.blue.offset = 0;
                info->var.transp.offset = 24;
                info->var.red.length = 8;
                info->var.green.length = 8;
                info->var.blue.length = 8;
                info->var.transp.length = 8;
                info->fix.visual = FB_VISUAL_TRUECOLOR;
                break;
        }
    
//        VIASETREG(par->io_base_virt, VIA_REG_CLIPTL, 0x0);
//        VIASETREG(par->io_base_virt, VIA_REG_CLIPBR, (info->var.yres_virtual << 16) | info->var.xres_virtual);  
        info->var.xoffset = 0;
        info->var.yoffset = 0;
        
        viafb_pan_display(&info->var,info);
        size = info->var.yres * info->var.xres * info->var.bits_per_pixel;
        memset((u8*)par->mem_base_virt, 0x00, size >> 3);
        return 0;
    }
    
    return -EINVAL; 
}

/**
 *  viafb_setcolreg - Optional function. Sets a color register.
 *  @regno: Which register in the CLUT we are programming 
 *  @red: The red value which can be up to 16 bits wide 
 *  @green: The green value which can be up to 16 bits wide 
 *  @blue:  The blue value which can be up to 16 bits wide.
 *  @transp: If supported the alpha value which can be up to 16 bits wide.  
 *  @info: frame buffer info structure
 * 
 *  Set a single color register. The values supplied have a 16 bit
 *  magnitude which needs to be scaled in this function for the hardware. 
 *  Things to take into consideration are how many color registers, if
 *  any, are supported with the current color visual. With truecolor mode
 *  no color palettes are supported. Here a psuedo palette is created 
 *  which we store the value in pseudo_palette in struct fb_info. For
 *  pseudocolor mode we have a limited color palette. To deal with this
 *  we can program what color is displayed for a particular pixel value.
 *  DirectColor is similar in that we can program each color field. If
 *  we have a static colormap we don't need to implement this function. 
 * 
 *  Returns negative errno on error, or zero on success.
 */
static int viafb_setcolreg(unsigned regno, unsigned red, unsigned green,
    unsigned blue, unsigned transp, struct fb_info *info)
{
    int bpp = info->var.bits_per_pixel;
    int m = (bpp==8)?256:16;
    u32 v;

    debug("set_colreg \n"); 
    debug("set_col_reg reg:%d r:%X g:%X b:%X t:%X bpp: %d\n",
        regno,red,green,blue,transp,bpp);
        
    if (regno >= m)
        return -EINVAL;
    
    palette[regno].red = red;
    palette[regno].green = green;
    palette[regno].blue = blue;
    palette[regno].transp = transp;

#if 0
    if (bpp==8) {
        t_outb(0xFF,0x3C6);
        t_outb(regno,0x3C8);
        t_outb(red>>10,0x3C9);
        t_outb(green>>10,0x3C9);
        t_outb(blue>>10,0x3C9);
    }
    else if (bpp == 16) {   // RGB 565 //
        ((u32*)info->pseudo_palette)[regno] = 
        (red & 0xF800) |
        ((green & 0xFC00) >> 5) | 
        ((blue & 0xF800) >> 11);
    }
    else if (bpp == 32) {   // ARGB 8888 //
        ((u32*)info->pseudo_palette)[regno] =
        ((transp & 0xFF00) <<16)  |
        ((red & 0xFF00) << 8)     |
        ((green & 0xFF00))        |
        ((blue & 0xFF00)>>8);
    }
#endif
      
#define CNVT_TOHW(val,width) ((((val)<<(width))+0x7FFF-(val))>>16)

    switch (info->fix.visual) {
        case FB_VISUAL_TRUECOLOR:
        case FB_VISUAL_PSEUDOCOLOR:
            red = CNVT_TOHW(red, info->var.red.length);
            green = CNVT_TOHW(green, info->var.green.length);
            blue = CNVT_TOHW(blue, info->var.blue.length);
            transp = CNVT_TOHW(transp, info->var.transp.length);
            break;
       case FB_VISUAL_DIRECTCOLOR:
            /* example here assumes 8 bit DAC. Might be different 
             * for your hardware */  
            red = CNVT_TOHW(red, 8);       
            green = CNVT_TOHW(green, 8);
            blue = CNVT_TOHW(blue, 8);
            /* hey, there is bug in transp handling... */
            transp = CNVT_TOHW(transp, 8);
            break;
    }

#undef CNVT_TOHW      

    v = (red << info->var.red.offset) |
        (green << info->var.green.offset) |
        (blue << info->var.blue.offset) |
        (transp << info->var.transp.offset);

    switch (info->var.bits_per_pixel) {
        case 8:
            t_outb(0xFF,0x3C6);
            t_outb(regno,0x3C8);
            t_outb(red,0x3C9);
            t_outb(green,0x3C9);
            t_outb(blue,0x3C9);
            break;  
        case 16:
            ((u32*)(info->pseudo_palette))[regno] = v;
            break;
        case 24:
        case 32:    
            ((u32*)(info->pseudo_palette))[regno] = v;
            break;
    }
    return 0;
}

static int viafb_pan_display(struct fb_var_screeninfo *var, struct fb_info *info)
{
    unsigned int offset;
    struct via_par * p = (struct via_par *)info->par;
    debug("pan_display\n");
    offset = (((var->yoffset * var->xres_virtual) + (var->xoffset))
        * var->bits_per_pixel / 8) & ~7;

//    printk("pan_display offset:%d %d\n",var->xoffset,var->yoffset);
    
    if ((var->xoffset) > (info->var.xres_virtual - info->var.xres)) {
        return -EINVAL;
    }
    if ((var->yoffset) > (info->var.yres_virtual - info->var.yres)) {
        return -EINVAL;
    }
    
    info->var.xoffset = var->xoffset;
    info->var.yoffset = var->yoffset;
            
//    printk("pan_display offset:%d\n",offset);
    
    if (var->activate & FB_ACTIVATE_VBL)
        write3X4(0x32, read3X4(0x32) & ~1);
    else
        write3X4(0x32, read3X4(0x32) | 1);
    
    VIASETREG(p->io_base_virt, VIA_REG_CRTCSTART, offset >> 1);
    return 0;
}

static ssize_t viafb_read(struct fb_info *info, char *buf, size_t count, loff_t *ppos)
{
    unsigned long   p   = *ppos;
    struct via_par *par = (struct via_par *)info->par;
    
    if (p >= par->video_mem_size)
        return 0;
    if (count >= par->video_mem_size)
        count = par->video_mem_size;
    if (count + p > par->video_mem_size)
        count = par->video_mem_size - p;
    
    viafb_sync(info);

    if (count) {
        char *base_addr;

        base_addr = info->screen_base;
        count -= copy_to_user(buf, base_addr+p, count);
        if (!count)
        return -EFAULT;
        *ppos += count;
    }
    return count;
}

static ssize_t viafb_write(struct fb_info *info, const char *buf, size_t count, loff_t *ppos)
{
    unsigned long   p   = *ppos;
    struct via_par *par = (struct via_par *)info->par;
    int err;

    if (p > par->video_mem_size)
        return -ENOSPC;
    if (count >= par->video_mem_size)
        count = par->video_mem_size;
    err = 0;
    if (count + p > par->video_mem_size) {
        count = par->video_mem_size - p;
        err = -ENOSPC;
    }
    
    viafb_sync(info);
    
    if (count) {
        char *base_addr;

        base_addr = info->screen_base;
        count -= copy_from_user(base_addr+p, buf, count);
        *ppos += count;
        err = -EFAULT;
    }

    if (count)
        return count;
    return err;
}

static int viafb_blank(int blank_mode, struct fb_info *info)
{
    unsigned char data;

    /* clear DPMS setting */
    data = read3X4(0x36);
    data &= 0xCF;
    
    switch (blank_mode) {
        case VESA_NO_BLANKING:
            /* Screen: On, HSync: On, VSync: On */
            data |= 0x00;
            break;
        case VESA_HSYNC_SUSPEND:
            /* Screen: Off, HSync: Off, VSync: On */
            data |= 0x10;
            break;
        case VESA_VSYNC_SUSPEND:
            /* Screen: Off, HSync: On, VSync: Off */
            data |= 0x20;
            break;
        case VESA_POWERDOWN:
            /* Screen: Off, HSync: Off, VSync: Off */
            data |= 0x30;
            break;
    }
    write3X4(0x36,data);
    return 0;
}

/*  modify from XServer/via_bios.c  */
int VIACheckTVExist(struct fb_info *info)
{
    struct via_par * par = (struct via_par *)info->par;

    I2C_Enable(I2CPort);

    switch (I2C_rdata(I2CPort, I2CAddrTV, 0x1B)) {
        case 2:
            par->TVEncoder = VIA_TV2PLUS;   // aka VT1621
            output("detected TV2PLUS (%d)\n", par->TVEncoder);
            break;
        case 3:
            par->TVEncoder = VIA_TV3;   // aka VT1622
            output("detected VT1622 (%d), using TV3 tables\n", par->TVEncoder);
            break;
        case 16:
            par->TVEncoder = VIA_VT1622A;
            output("detected VT1622A (%d)\n", par->TVEncoder);
            break;
        default:
            par->TVEncoder = VIA_NONETV;
            break;
    }
    return par->TVEncoder;
}

/*  modify from XServer/via_bios.c  */
void VIAPreSetTV3Mode(struct fb_info *info,int xres,int yres,int bpp)
{
    struct via_par * par = (struct via_par *)info->par;
    
    VIABIOSTVMASKTablePtr TVMaskTbl;
    unsigned char *TV;
    unsigned int *DotCrawl, *RGB, *YCbCr, *Patch2;
    int tvIndx, tvType;
    int i, j;
    unsigned char  W_Buffer[VIA_BIOS_MAX_NUM_TV_REG];
    int w_bytes;

    debug("VIAPreSetTV3Mode x:%d y:%d bpp:%d\n",xres,yres,bpp);
    
    DotCrawl = NULL;
    RGB = NULL;
    YCbCr = NULL;
    Patch2 = NULL;

    if (xres==640 && yres==480)
        par->TVindex = VIA_TVRES_640X480;
    else if (xres==800 && yres==600)
        par->TVindex = VIA_TVRES_800X600;
    else if (xres==1024 && yres==768)
        par->TVindex = VIA_TVRES_1024X768;
    else if (xres==848 && yres==480)
        par->TVindex = VIA_TVRES_848X480;
    else if (xres==720 && yres==576) {
        par->TVindex = VIA_TVRES_720X576;
        par->TVtype = TVTYPE_PAL;             /* PAL only */
    } else if (xres==720 && yres==480) {
        par->TVindex = VIA_TVRES_720X480;
        par->TVtype = TVTYPE_NTSC;            /* NTSC only */
    } else {
        debug("No support TV mode:%dx%d\n",xres,yres);
        return;
    }
    tvIndx = par->TVindex;
    tvType = par->TVtype;

    TV = NULL;
    W_Buffer[0] = 0;
    for (i = 0; i < VIA_BIOS_MAX_NUM_TV_REG; i++) {
        W_Buffer[i+1] = 0;
    }

    if (par->TVEncoder == VIA_TV3) {
        TVMaskTbl = &tv3MaskTable;
        if (tvType == TVTYPE_PAL) {
            switch (par->TVoverscan) {
                case VIA_TVNORMAL:
                    TV = tv3Table[tvIndx].TVPAL;
                    RGB = tv3Table[tvIndx].RGBPAL;
                    YCbCr = tv3Table[tvIndx].YCbCrPAL;
                    Patch2 = tv3Table[tvIndx].PatchPAL2;
                    break;
                case VIA_TVOVER:
                    TV = tv3OverTable[tvIndx].TVPAL;
                    RGB = tv3OverTable[tvIndx].RGBPAL;
                    YCbCr = tv3OverTable[tvIndx].YCbCrPAL;
                    Patch2 = tv3OverTable[tvIndx].PatchPAL2;
                    break;
            }
        }
        else {
            switch (par->TVoverscan) {
                case VIA_TVNORMAL:
                    TV = tv3Table[tvIndx].TVNTSC;
                    DotCrawl = tv3Table[tvIndx].DotCrawlNTSC;
                    RGB = tv3Table[tvIndx].RGBNTSC;
                    YCbCr = tv3Table[tvIndx].YCbCrNTSC;
                    Patch2 = tv3Table[tvIndx].PatchNTSC2;
                    break;
                case VIA_TVOVER:
                    TV = tv3OverTable[tvIndx].TVNTSC;
                    DotCrawl = tv3OverTable[tvIndx].DotCrawlNTSC;
                    RGB = tv3OverTable[tvIndx].RGBNTSC;
                    YCbCr = tv3OverTable[tvIndx].YCbCrNTSC;
                    Patch2 = tv3OverTable[tvIndx].PatchNTSC2;
                    break;
            }
        }
    }
    else {
        TVMaskTbl = &vt1622aMaskTable;
        if (tvType == TVTYPE_PAL) {
            switch (par->TVoverscan) {
                case VIA_TVNORMAL:
                    TV = vt1622aTable[tvIndx].TVPAL;
                    RGB = vt1622aTable[tvIndx].RGBPAL;
                    YCbCr = vt1622aTable[tvIndx].YCbCrPAL;
                    Patch2 = vt1622aTable[tvIndx].PatchPAL2;
                    break;
                case VIA_TVOVER:
                    TV = vt1622aOverTable[tvIndx].TVPAL;
                    RGB = vt1622aOverTable[tvIndx].RGBPAL;
                    YCbCr = vt1622aOverTable[tvIndx].YCbCrPAL;
                    Patch2 = vt1622aOverTable[tvIndx].PatchPAL2;
                    break;
            }
        }
        else {
            switch (par->TVoverscan) {
                case VIA_TVNORMAL:
                    TV = vt1622aTable[tvIndx].TVNTSC;
                    DotCrawl = vt1622aTable[tvIndx].DotCrawlNTSC;
                    RGB = vt1622aTable[tvIndx].RGBNTSC;
                    YCbCr = vt1622aTable[tvIndx].YCbCrNTSC;
                    Patch2 = vt1622aTable[tvIndx].PatchNTSC2;
                    break;
                case VIA_TVOVER:
                    TV = vt1622aOverTable[tvIndx].TVNTSC;
                    DotCrawl = vt1622aOverTable[tvIndx].DotCrawlNTSC;
                    RGB = vt1622aOverTable[tvIndx].RGBNTSC;
                    YCbCr = vt1622aOverTable[tvIndx].YCbCrNTSC;
                    Patch2 = vt1622aOverTable[tvIndx].PatchNTSC2;
                    break;
            }
        }
    }

    for (i = 0, j = 0; i < TVMaskTbl->numTV; j++) {
        if (TVMaskTbl->TV[j] == 0xFF) {
            W_Buffer[j+1] = TV[j];
            i++;
        }
    }

    w_bytes = j + 1;

    I2C_Enable(I2CPort);

    /* TV Reset */
    I2C_wdata(I2CPort, I2CAddrTV, 0x1D, 0);
    /* Wait for next vsync and then reset interrupt count --
     * we should then get the correct field parity */
    viafb_wait_for_sync(info, VIAFB_WAIT_ANY);
    msleep(2);  // allow for vblank-vsync interval
    par->irq_cnt = 0;
    /* Clear reset */
    I2C_wdata(I2CPort, I2CAddrTV, 0x1D, 0x80);

    I2C_Write(I2CPort, I2CAddrTV, W_Buffer, w_bytes);

    /* Turn on all Composite and S-Video output */
    I2C_wdata(I2CPort, I2CAddrTV, 0x0E, 0);
}

void VIAPostSetTV3Mode(struct fb_info *info,int xres,int yres,int bpp)
{
    struct via_par * par = (struct via_par *)info->par;
    VIABIOSTVMASKTablePtr TVMaskTbl;
    unsigned char   *CRTC1, *CRTC2, *Misc1, *Misc2;
    int             tvIndx, tvType;
    int             i, j, data;

    debug("VIAPostSetTV3Mode x:%d y:%d bpp:%d\n",xres,yres,bpp);
    
    CRTC1 = NULL;
    CRTC2 = NULL;
    Misc1 = NULL;
    Misc2 = NULL;

    tvIndx = par->TVindex;
    tvType = par->TVtype;

    if (par->TVEncoder == VIA_TV3) {
        TVMaskTbl = &tv3MaskTable;
        if (tvType == TVTYPE_PAL) {
            switch (par->TVoverscan) {
                case VIA_TVNORMAL:
                    CRTC1 = tv3Table[tvIndx].CRTCPAL1;
                    switch (bpp) {
                        case 8:
                            CRTC2 = tv3Table[tvIndx].CRTCPAL2_8BPP;
                            break;
                        case 16:
                            CRTC2 = tv3Table[tvIndx].CRTCPAL2_16BPP;
                            break;
                        case 24:
                        case 32:
                            CRTC2 = tv3Table[tvIndx].CRTCPAL2_32BPP;
                            break;
                    }
                    Misc1 = tv3Table[tvIndx].MiscPAL1;
                    Misc2 = tv3Table[tvIndx].MiscPAL2;
                    break;
                case VIA_TVOVER:
                    CRTC1 = tv3OverTable[tvIndx].CRTCPAL1;
                    switch (bpp) {
                        case 8:
                            CRTC2 = tv3OverTable[tvIndx].CRTCPAL2_8BPP;
                            break;
                        case 16:
                            CRTC2 = tv3OverTable[tvIndx].CRTCPAL2_16BPP;
                            break;
                        case 24:
                        case 32:
                            CRTC2 = tv3OverTable[tvIndx].CRTCPAL2_32BPP;
                            break;
                    }
                    Misc1 = tv3OverTable[tvIndx].MiscPAL1;
                    Misc2 = tv3OverTable[tvIndx].MiscPAL2;
                    break;
            }
        }
        else {
            switch (par->TVoverscan) {
                case VIA_TVNORMAL:
                    CRTC1 = tv3Table[tvIndx].CRTCNTSC1;
                    switch (bpp) {
                        case 8:
                            CRTC2 = tv3Table[tvIndx].CRTCNTSC2_8BPP;
                            break;
                        case 16:
                            CRTC2 = tv3Table[tvIndx].CRTCNTSC2_16BPP;
                            break;
                        case 24:
                        case 32:
                            CRTC2 = tv3Table[tvIndx].CRTCNTSC2_32BPP;
                            break;
                    }
                    Misc1 = tv3Table[tvIndx].MiscNTSC1;
                    Misc2 = tv3Table[tvIndx].MiscNTSC2;
                    break;
                case VIA_TVOVER:
                    CRTC1 = tv3OverTable[tvIndx].CRTCNTSC1;
                    switch (bpp) {
                        case 8:
                            CRTC2 = tv3OverTable[tvIndx].CRTCNTSC2_8BPP;
                            break;
                        case 16:
                            CRTC2 = tv3OverTable[tvIndx].CRTCNTSC2_16BPP;
                            break;
                        case 24:
                        case 32:
                            CRTC2 = tv3OverTable[tvIndx].CRTCNTSC2_32BPP;
                            break;
                    }
                    Misc1 = tv3OverTable[tvIndx].MiscNTSC1;
                    Misc2 = tv3OverTable[tvIndx].MiscNTSC2;
                    break;
            }
        }
    }
    else {
        TVMaskTbl = &vt1622aMaskTable;
        if (tvType == TVTYPE_PAL) {
            switch (par->TVoverscan) {
                case VIA_TVNORMAL:
                    CRTC1 = vt1622aTable[tvIndx].CRTCPAL1;
                    switch (bpp) {
                        case 8:
                            CRTC2 = vt1622aTable[tvIndx].CRTCPAL2_8BPP;
                            break;
                        case 16:
                            CRTC2 = vt1622aTable[tvIndx].CRTCPAL2_16BPP;
                            break;
                        case 24:
                        case 32:
                            CRTC2 = vt1622aTable[tvIndx].CRTCPAL2_32BPP;
                            break;
                    }
                    Misc1 = vt1622aTable[tvIndx].MiscPAL1;
                    Misc2 = vt1622aTable[tvIndx].MiscPAL2;
                    break;
                case VIA_TVOVER:
                    CRTC1 = vt1622aOverTable[tvIndx].CRTCPAL1;
                    switch (bpp) {
                        case 8:
                            CRTC2 = vt1622aOverTable[tvIndx].CRTCPAL2_8BPP;
                            break;
                        case 16:
                            CRTC2 = vt1622aOverTable[tvIndx].CRTCPAL2_16BPP;
                            break;
                        case 24:
                        case 32:
                            CRTC2 = vt1622aOverTable[tvIndx].CRTCPAL2_32BPP;
                            break;
                    }
                    Misc1 = vt1622aOverTable[tvIndx].MiscPAL1;
                    Misc2 = vt1622aOverTable[tvIndx].MiscPAL2;
                    break;
            }
        }
        else {
            switch (par->TVoverscan) {
                case VIA_TVNORMAL:
                    CRTC1 = vt1622aTable[tvIndx].CRTCNTSC1;
                    switch (bpp) {
                        case 8:
                            CRTC2 = vt1622aTable[tvIndx].CRTCNTSC2_8BPP;
                            break;
                        case 16:
                            CRTC2 = vt1622aTable[tvIndx].CRTCNTSC2_16BPP;
                            break;
                        case 24:
                        case 32:
                            CRTC2 = vt1622aTable[tvIndx].CRTCNTSC2_32BPP;
                            break;
                    }
                    Misc1 = vt1622aTable[tvIndx].MiscNTSC1;
                    Misc2 = vt1622aTable[tvIndx].MiscNTSC2;
                    break;
                case VIA_TVOVER:
                    CRTC1 = vt1622aOverTable[tvIndx].CRTCNTSC1;
                    switch (bpp) {
                        case 8:
                            CRTC2 = vt1622aOverTable[tvIndx].CRTCNTSC2_8BPP;
                            break;
                        case 16:
                            CRTC2 = vt1622aOverTable[tvIndx].CRTCNTSC2_16BPP;
                            break;
                        case 24:
                        case 32:
                            CRTC2 = vt1622aOverTable[tvIndx].CRTCNTSC2_32BPP;
                            break;
                    }
                    Misc1 = vt1622aOverTable[tvIndx].MiscNTSC1;
                    Misc2 = vt1622aOverTable[tvIndx].MiscNTSC2;
                    break;
            }
        }
    }

    for (i = 0, j = 0; i < TVMaskTbl->numCRTC1; j++) {
        if (TVMaskTbl->CRTC1[j] == 0xFF) {
            write3X4(j, CRTC1[j]);
            i++;
        }
    }

    j = 0;

    if (Misc1[j] & 0x20) {
        write3X4(0x33, (read3X4(0x33) | 0x20));
        j++;
    }
    else {
        write3X4(0x33, (read3X4(0x33) & 0xdf));
        j++;
    }
    write3X4(0x6A, Misc1[j++]);
    write3X4(0x6B, Misc1[j++] | 0x01);

    write3X4(0x6C, (Misc1[j++] & 0xE1) | 0x01);

    if (TVMaskTbl->misc1 & 0x30) {
        write3X4(0x17, read3X4(0x17) & 0x7F);

        write3C4(0x46, Misc1[j++]);
        write3C4(0x47, Misc1[j++]);

        write3X4(0x17, read3X4(0x17) | 0x80);

        write3C4(0x40, read3C4(0x40) | 0x02);
        write3C4(0x40, read3C4(0x40) & 0xFD);

        /* Use external clock */
        data = t_inb(0x3cc) | 0x0C;
        t_outb(data, 0x3c2);
    }
}

static int viafb_sync(struct fb_info *info)
{
    WaitIdleCLE266(info);
    return 0;
}

static int viafb_wait_for_sync(struct fb_info *info, int field_option)
{
    wait_queue_t __wait;
    unsigned int cnt;
    int ret=0;
    struct via_par * p = (struct via_par *)info->par;
    
    init_waitqueue_entry(&__wait, current);

    cnt = p->irq_cnt;
    
    if (field_option == VIAFB_WAIT_ANY) // field sync
    {
        ret = wait_event_interruptible_timeout(p->wq, cnt != p->irq_cnt, HZ/10);
    }
    else if (field_option == VIAFB_WAIT_TOPFIELD) // frame sync
    {
        ret = wait_event_interruptible_timeout(p->wq, (cnt != p->irq_cnt) && ((p->irq_cnt & 1)==0), HZ/10);
    }
    else if (field_option == VIAFB_WAIT_BOTTOMFIELD) // wait for bottom field
    {
        ret = wait_event_interruptible_timeout(p->wq, (cnt != p->irq_cnt) && ((p->irq_cnt & 1)==1), HZ/10);
    }
    else if (field_option == VIAFB_WAIT_FLIP) // wait for any pending flip to complete
    {
        ret = wait_event_interruptible_timeout(p->wq, p->flip_pending == 0, HZ/10);
    }

    if (ret < 0) {
        return ret;
    }
    if (ret == 0) {
        return -ETIMEDOUT;
    }

    return 0;
}

static int viafb_flip_on_sync(struct fb_info *info, struct fb_flip *flip)
{
    struct via_par * p = (struct via_par *)info->par;
    
    if (flip->device == VIAFB_FLIP_V1)
    {
        p->flip = *flip;
        flip->field = p->irq_cnt % 2;
        flip->count = p->irq_cnt / 2;
        p->flip_pending = 1;
        return 0;
    }
    else if (flip->device == VIAFB_FLIP_NOP)
    {
        return 0;
    }
    return -EINVAL;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static int viafb_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
    unsigned long arg, struct fb_info *info)
#else
static int viafb_ioctl(struct fb_info *info, unsigned int cmd, unsigned long arg)
#endif
{
    switch (cmd) {
        case FBIO_WAITFORVSYNC:
        {
            u_int32_t field_option;
            
            if (get_user(field_option, (u_int32_t __user *)arg))
                return -EFAULT;
            
            return viafb_wait_for_sync(info, field_option);
        }
        case FBIO_FLIPONVSYNC:
        {
            struct fb_flip flip;
            int err;

            if (copy_from_user(&flip, (void __user *)arg, sizeof(flip)))
                return -EFAULT;

            err = viafb_flip_on_sync(info, &flip);
            if (err)
                return err;

            if (copy_to_user((void __user *)arg, &flip, sizeof(flip)))
                return -EFAULT;

            return 0;
        }
    }

    return -ENOTTY;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static irqreturn_t via_vga_irqhandler(int irq, void *ptr, struct pt_regs *regs)
#else
static irqreturn_t via_vga_irqhandler(int irq, void *ptr)
#endif
{
    struct fb_info *info = (struct fb_info *)ptr;
    struct via_par * p = (struct via_par *)info->par;
    int handled = 0;
    
    u_int32_t status = VIAGETREG(p->io_base_virt, VIA_REG_INTERRUPT);

    if (status & VIA_IRQ_VBI_PENDING) {
        VIASETREG(p->io_base_virt, VIA_REG_INTERRUPT, status | VIA_IRQ_VBI_PENDING);

        p->irq_cnt++;
        if (p->flip_pending && ((p->irq_cnt & 1) == p->flip.field))
        {
            VIASETREG(p->io_base_virt, V1_STARTADDR_0, p->flip.offset[0]);
            VIASETREG(p->io_base_virt, V1_STARTADDR_CB0, p->flip.offset[1]);
            VIASETREG(p->io_base_virt, V1_STARTADDR_CR0, p->flip.offset[2]);
            VIASETREG(p->io_base_virt, V_COMPOSE_MODE, VIAGETREG(p->io_base_virt, V_COMPOSE_MODE)|V1_COMMAND_FIRE);
            p->flip_pending = 0;
        }
        wake_up_interruptible(&p->wq);
        handled = 1;
    }
    return IRQ_RETVAL(handled);
}

static int install_irq_handler(struct fb_info *info, int irq_num)
{
    int retval;
    struct via_par * p = (struct via_par *)info->par;
    
    retval = request_irq(irq_num, via_vga_irqhandler,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
        SA_SHIRQ | SA_INTERRUPT, "viafb", info);
#else
        IRQF_SHARED | IRQF_DISABLED, "viafb", info);
#endif

    if (retval == -EINVAL) {
        output("No IRQ handler installed, IRQ number %d is not valid.\n", irq_num);
    }
    else if (retval == -EBUSY) {
        output("No IRQ handler installed, IRQ %d is busy. Check BIOS.\n", irq_num);
    }
    else if (retval < 0) {
        output("No IRQ handler installed. Tried IRQ number %d.\n", irq_num);
    }

    p->irq_cnt = 0;
    if (retval >= 0) {
        p->irq_num = irq_num;

        // Enable vertical sync IRQ.
        // BUG WARNING: May conflict with other VIA drivers using vsync IRQ.

        VIASETREG(p->io_base_virt, VIA_REG_INTERRUPT,
            VIAGETREG(p->io_base_virt, VIA_REG_INTERRUPT) |
            VIA_IRQ_GLOBAL | VIA_IRQ_VBI_ENABLE | VIA_IRQ_VBI_PENDING);

        outb(0x11, 0x3d4);
        outb(inb(0x3d5) | 0x30, 0x3d5);

        output("irq handler installed, IRQ(0x200) = %x\n", VIAGETREG(p->io_base_virt, VIA_REG_INTERRUPT));
    }

    return retval;
}

static void uninstall_irq_handler(struct fb_info *info)
{
    // Disable vertical sync IRQ.
    // BUG WARNING: May conflict with other VIA drivers using vsync IRQ.

    struct via_par * p = (struct via_par *)info->par;
    if (p->irq_num == 0)
        return;
    
    VIASETREG(p->io_base_virt, VIA_REG_INTERRUPT,
        (VIAGETREG(p->io_base_virt, VIA_REG_INTERRUPT) & ~VIA_IRQ_VBI_ENABLE)
        | VIA_IRQ_VBI_PENDING);

    outb(0x11, 0x3d4);
    outb(inb(0x3d5) & ~0x30, 0x3d5);

    free_irq(p->irq_num, info);
    output("%d interrupt requests serviced.\n", p->irq_cnt);
}

static struct fb_ops viafb_ops = {
    .owner          = THIS_MODULE,
    .fb_open        = viafb_open,
    .fb_release     = viafb_release,
    .fb_check_var   = viafb_check_var,
    .fb_set_par     = viafb_set_par,    
    .fb_setcolreg   = viafb_setcolreg,
    .fb_blank       = viafb_blank,
    .fb_pan_display = viafb_pan_display,    
    .fb_fillrect    = viafb_fillrect,
    .fb_copyarea    = viafb_copyarea,
    .fb_imageblit   = viafb_imageblit,
    .fb_cursor      = viafb_cursor,
    .fb_ioctl       = viafb_ioctl,
    .fb_sync        = viafb_sync,
    .fb_read        = viafb_read,
    .fb_write       = viafb_write,
};

static int __devinit via_pci_probe(struct pci_dev * dev, const struct pci_device_id * id)
{
    int cmap_len = 256;
    int i;
    unsigned int default_xres,default_yres;
    char *tmpc,*tmpm;
    u8 rev;
    
    info=framebuffer_alloc(sizeof(struct via_par), NULL);
    current_par = (struct via_par*)info->par;
    
    // Map physical IO memory address into kernel space.
    current_par->io_base_phy = pci_resource_start(dev, 1);
    debug("io_base_phy : %X\n",current_par->io_base_phy);
    current_par->io_size = pci_resource_len(dev, 1);
    debug("io_size     : %X\n",current_par->io_size);
    
    via_enable_mmio();
    
    /* unlock extended register */
    write3C4(0x10, 1);
    
    if (!request_mem_region(current_par->io_base_phy, current_par->io_size, "viafb")) {
        output("request_mem_region for mmio failed!\n");
        return -1;
    }
    
    current_par->io_base_virt = (unsigned int)ioremap(current_par->io_base_phy, current_par->io_size);
    if (!current_par->io_base_virt) {
        release_mem_region(current_par->io_base_phy, current_par->io_size);
        output("ioremap for mmio failed\n");
        return -1;
    }
    
    output("viafb : %X\n",current_par->io_base_virt);
    debug("viafb : %X\n",current_par->io_size);
    viafb_fix.mmio_start = current_par->io_base_phy;
    debug("viafb_fix.mmio_start : %X\n",viafb_fix.mmio_start);
    
    viafb_fix.mmio_len = current_par->io_size;
    
    current_par->mem_base_phy = pci_resource_start(dev,0);
    debug("mem_base_phy  : %X\n",current_par->mem_base_phy);
    current_par->mem_size = get_memsize();
    debug("mem_size : %X\n",current_par->mem_size);
    
    if (!request_mem_region(current_par->mem_base_phy, current_par->mem_size, "viafb")) {
        iounmap((void *)current_par->io_base_virt);
        release_mem_region(current_par->io_base_phy, current_par->io_size);
        output("request_mem_region for framebuffer failed!\n");
        return -1;
    }

    viafb_fix.smem_start = current_par->mem_base_phy;
    current_par->video_mem_size = current_par->mem_size;
    
    current_par->mem_base_virt = (unsigned int)ioremap(current_par->mem_base_phy, current_par->mem_size);
    if (!current_par->mem_base_virt) {
        release_mem_region(current_par->mem_base_phy, current_par->mem_size);
        iounmap((void *)current_par->io_base_virt);
        release_mem_region(current_par->io_base_phy, current_par->io_size);
        output("ioremap for framebuffer failed\n");
        return -1;
    }

    info->screen_base = (u8*)current_par->mem_base_virt;
    
    pci_read_config_byte(dev, 0xf6, &rev);
    output("Found Device Rev:%d\n", rev);
    
    mtrr_del( 4, current_par->mem_base_phy, 0x800000 );
    
    /* added by andi */
    mtrr_add( current_par->mem_base_phy, current_par->mem_size, MTRR_TYPE_WRCOMB, 1 );
    
    info->fbops = &viafb_ops;
    info->fix = viafb_fix;
    info->pseudo_palette = pseudo_pal;
    info->flags = FBINFO_DEFAULT;

    tmpm = mode;
    tmpc = strsep(&tmpm,"x");
    default_xres = simple_strtoul(tmpc,NULL,0);
    default_yres = simple_strtoul(tmpm,NULL,0);
    output("X:%d Y:%d\n",default_xres,default_yres);
    
    info->fix.accel = FB_ACCEL_VIA_UNICHROME;
    info->fix.line_length = default_xres * bpp / 8;
    info->var.xres = default_xres;  /* visible resolution */
    info->var.yres = default_yres;
    info->var.bits_per_pixel = bpp;
    info->var.xres_virtual = default_xres;  /* virtual resolution */
    info->var.yres_virtual = default_yres * 2;
    info->var.xoffset = 0;  /* offset from virtual to ... */
    info->var.yoffset = 0;  /* ... visible resolution */
    
    switch (info->var.bits_per_pixel) {
        case 8:
            info->var.red.offset = 0;
            info->var.green.offset = 0;
            info->var.blue.offset = 0;
            info->var.red.length = 6;
            info->var.green.length = 6;
            info->var.blue.length = 6;
            info->fix.visual = FB_VISUAL_PSEUDOCOLOR;
            break;
        case 16:
            info->var.red.offset = 11;
            info->var.green.offset = 5;
            info->var.blue.offset = 0;
            info->var.red.length = 5;
            info->var.green.length = 6;
            info->var.blue.length = 5;
            info->fix.visual = FB_VISUAL_TRUECOLOR;
            break;
        case 32:
            info->var.red.offset = 16;
            info->var.green.offset = 8;
            info->var.blue.offset = 0;
            info->var.transp.offset = 24;
            info->var.red.length = 8;
            info->var.green.length = 8;
            info->var.blue.length = 8;
            info->var.transp.length = 8;
            info->fix.visual = FB_VISUAL_TRUECOLOR;
            break;
    }
    
    info->var.pixclock=10000000/info->var.xres*1000/info->var.yres;
    info->var.left_margin=(info->var.xres>>3)&0xf8;
    info->var.right_margin=32;
    info->var.upper_margin=16;
    info->var.lower_margin=4;
    info->var.hsync_len=info->var.left_margin;
    info->var.vsync_len=4;
    info->var.accel_flags = 0;
    info->var.activate = FB_ACTIVATE_NOW;
    
    current_par->noaccel = noaccel;
    current_par->refresh = refresh;
    current_par->TVon= TVon;
    current_par->TVtype = TVtype;
    current_par->TVoverscan = TVoverscan;
    current_par->TVEncoder = TVEncoder;
    
    current_par->flip_pending = 0;
    
    if (TVon) {
        VIACheckTVExist(info);
        if (default_xres==640 && default_yres==480)
            current_par->TVindex = VIA_TVRES_640X480;
        else if (default_xres==800 && default_yres==600)
            current_par->TVindex = VIA_TVRES_800X600;
        else if (default_xres==1024 && default_yres==768)
            current_par->TVindex = VIA_TVRES_1024X768;
        else if (default_xres==848 && default_yres==480)
            current_par->TVindex = VIA_TVRES_848X480;
        else if (default_xres==720 && default_yres==576) {
            current_par->TVindex = VIA_TVRES_720X576;
            current_par->TVtype = TVTYPE_PAL;             /* PAL only */
        }
        else if (default_xres==720 && default_yres==480) {
            current_par->TVindex = VIA_TVRES_720X480;
            current_par->TVtype = TVTYPE_NTSC;            /* NTSC only */
        }
        else {
            debug("No support TV mode:%dx%d\n",default_xres,default_yres);
            current_par->TVon = 0;
        }
    }
    if (current_par->TVtype == TVTYPE_NONE)
        current_par->TVon = 0;
    if (!current_par->TVon) {
        if (current_par->refresh != 0xFF) {
            for (i=0 ; refreshTable[resMode][i].refresh != 0 && current_par->refresh >= refreshTable[resMode][i].refresh ; i++) {
                // nothing
            }
            if (i==0)
                current_par->refresh = 0xFF;
            else
                current_par->refresh = refreshTable[resMode][i-1].refresh;
        }
    }
    else {
        if (current_par->TVtype == TVTYPE_NTSC)
            current_par->refresh = 60;
        else // TVtype == TVTYPE_PAL
            current_par->refresh = 50;
    }
    output("mode=%s  bpp=%d  refresh=%d  TVon=%d  TVtype=%d \n",mode,bpp,refresh,TVon,TVtype);
    
    current_par->vq_start = current_par->video_mem_size - VIA_VQ_SIZE;
    current_par->vq_end   = current_par->vq_start + VIA_VQ_SIZE - 1;
    current_par->vq_size  = VIA_VQ_SIZE;//info.par->mem_size;
    
    current_par->video_mem_size -= VIA_VQ_SIZE;
    
    output("VQ start:%X  end:%X  size:%X\n",current_par->vq_start,current_par->vq_end,current_par->vq_size);
//    memset((u8*)current_par->vq_start,0x00,current_par->vq_size);
    
    current_par->cur_start = current_par->video_mem_size - VIA_CURSOR_SIZE;
    current_par->cur_end   = current_par->cur_start + VIA_CURSOR_SIZE - 1;
    current_par->cur_size  = VIA_CURSOR_SIZE;//info.par->mem_size;
    
    current_par->video_mem_size -= VIA_CURSOR_SIZE;
    output("Cursor start:%X  end:%X  size:%X\n",current_par->cur_start,current_par->cur_end,current_par->cur_size);
    
    /* set the available videomemory */
    info->fix.smem_len = current_par->video_mem_size;
    
    if (current_par->TVon) {
        VIACheckTVExist(info);
        if (default_xres==640 && default_yres==480)
            current_par->TVindex = VIA_TVRES_640X480;
        else if (default_xres==800 && default_yres==600)
            current_par->TVindex = VIA_TVRES_800X600;
        else if (default_xres==1024 && default_yres==768)
            current_par->TVindex = VIA_TVRES_1024X768;
        else if (default_xres==848 && default_yres==480)
            current_par->TVindex = VIA_TVRES_848X480;
        else if (default_xres==720 && default_yres==576) {
            current_par->TVindex = VIA_TVRES_720X576;
            current_par->TVtype = TVTYPE_PAL;             /* PAL only */
        }
        else if (default_xres==720 && default_yres==480) {
            current_par->TVindex = VIA_TVRES_720X480;
            current_par->TVtype = TVTYPE_NTSC;            /* NTSC only */
        }
        else {
            debug("No support TV mode:%dx%d\n",default_xres,default_yres);
            current_par->TVon = 0;
        }
    }
    
    if (current_par->TVtype == TVTYPE_NONE)
        current_par->TVon = 0;
    
    if (!current_par->TVon) {
        if (current_par->refresh != 0xFF) {
            for (i=0 ; refreshTable[resMode][i].refresh != 0 && current_par->refresh >= refreshTable[resMode][i].refresh ; i++) {
                // nothing
            }
            if (i==0)
                current_par->refresh = 0xFF;
            else
                current_par->refresh = refreshTable[resMode][i-1].refresh;
        }
    }
    else {
        if (current_par->TVtype == TVTYPE_NTSC)
            current_par->refresh = 60;
        else // TVtype == TVTYPE_PAL
            current_par->refresh = 50;
    }
    
    output("mode=%s  bpp=%d  refresh=%d  TVon=%d  TVtype=%d \n",
        mode,bpp,current_par->refresh,current_par->TVon,current_par->TVtype);    
    
    
    init_waitqueue_head(&current_par->wq);
    
    // Try to install a vertical blanking interrupt handler.
    // Failure is not fatal.
    install_irq_handler(info, dev->irq);
            

    fb_alloc_cmap(&info->cmap, cmap_len, 0);
    if (register_framebuffer(info) < 0)
    {
        uninstall_irq_handler(info);
	fb_dealloc_cmap(&info->cmap);
        return -EINVAL;
    }

    if (setmode(info,info->var.xres,info->var.yres,info->var.bits_per_pixel) < 0)
    {
      uninstall_irq_handler(info);
      fb_dealloc_cmap(&info->cmap);
      return -EINVAL;
    }
    viafb_pan_display(&info->var,info);
    
    info->var.pixclock=10000000/info->var.xres*1000/info->var.yres;
    info->var.left_margin=(info->var.xres>>3)&0xf8;
    info->var.right_margin=32;
    info->var.upper_margin=16;
    info->var.lower_margin=4;
    info->var.hsync_len=info->var.left_margin;
    info->var.vsync_len=4;
    
    viafb_init_2d_engine(info, 1);
    viafb_init_cursor(info);
    
    
    printk(KERN_INFO "fb%d: %s frame buffer device\n", info->node, info->fix.id);
    return 0;
}

static void __devexit via_pci_remove(struct pci_dev * dev)
{
    uninstall_irq_handler(info);
    unregister_framebuffer(info);
    
    via_disable_mmio();
    iounmap((void *)current_par->mem_base_virt);
    iounmap((void *)current_par->io_base_virt);
    release_mem_region(current_par->mem_base_phy, current_par->mem_size);
    release_mem_region(current_par->io_base_phy, current_par->io_size);
    
    framebuffer_release(info);
}

/* List of boards that we are trying to support */
static struct pci_device_id via_devices[] __devinitdata = {
    {PCI_VENDOR_ID_VIA, 0x3122, PCI_ANY_ID,PCI_ANY_ID,0,0,0},
    {PCI_VENDOR_ID_VIA, 0x7205, PCI_ANY_ID,PCI_ANY_ID,0,0,0},
    {0,}
};
    
MODULE_DEVICE_TABLE(pci,via_devices);

static struct pci_driver viafb_pci_driver = {
    name:"viafb",
    id_table:via_devices,
    probe:via_pci_probe,
    remove:__devexit_p(via_pci_remove),
};

int __init viafb_init(void)
{
#ifndef MODULE
    char *option = NULL;

    if (fb_get_options("viafb", &option))
        return -ENODEV;
    viafb_setup(option);
#endif

    output("VIA UNICHROME framebuffer %s initializing\n", VERSION);
    return pci_register_driver(&viafb_pci_driver);
}

static void __exit viafb_cleanup(void)
{
    output("VIA UNICHROME framebuffer %s cleanup\n", VERSION); 
    pci_unregister_driver(&viafb_pci_driver);
}

#ifndef MODULE
static int viafb_setup(char *options)
{
    /* Parse user speficied options (`video=viafb:') */
    char * opt;
    
//    printk("opt: %s\n",options);

    if (!options || !*options)
        return 0;

    while((opt = strsep(&options,",")) != NULL ) {
        if (!*opt)
            continue;
        
        if (!strncmp(opt,"noaccel",7))
            noaccel = 1;
        else if (!strncmp(opt,"bpp=",4))
            bpp = simple_strtoul(opt+4,NULL,0);
        else if (!strncmp(opt,"TVon=",5))
            TVon = simple_strtoul(opt+5,NULL,0);
        else if (!strncmp(opt,"TVtype=",7))
            TVtype = simple_strtoul(opt+7,NULL,0);
        else if (!strncmp(opt,"TVoverscan=",11))
            TVoverscan = simple_strtoul(opt+11,NULL,0);
        else if (!strncmp(opt,"memsize=",8))
            memsize = simple_strtoul(opt+8,NULL,0);
        else if (!strncmp(opt,"refresh=",8))
            refresh = simple_strtoul(opt+8,NULL,0);
        else if (!strncmp(opt, "mode=", 5))
            mode = opt+5;
        else
            mode = opt;
    }
    return 0;
}
#endif
#ifdef MODULE

module_param(mode, charp, S_IRUGO);
module_param(bpp, int, S_IRUGO);
module_param(refresh, int, S_IRUGO);
module_param(noaccel, int, S_IRUGO);
module_param(TVon,int, S_IRUGO);        //  0:off   1:on
module_param(TVtype,int , S_IRUGO);     //  0:none  1:NTSC  2:PAL
module_param(TVoverscan,int ,S_IRUGO);  //  0:off   1:on
module_param(memsize,int, S_IRUGO);

MODULE_AUTHOR("Andre Kriehn");
MODULE_DESCRIPTION("UNICHROME FB-Driver");
MODULE_LICENSE("GPL");

#endif

module_init(viafb_init);
module_exit(viafb_cleanup);
