/*
 * Copyright (C) 2003 by [jyhmingfeng@via.com.tw]
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU *General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) any later 
 * version. 
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - 
 * Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef VIAFB_H
#define VIAFB_H

#ifndef VIAFB_DEBUG
#define VIAFB_DEBUG 0
#endif

#if VIAFB_DEBUG
#define debug(f,a...) printk("%s:" f,  __FUNCTION__ , ## a)
#else
#define debug(f,a...)
#endif

#define output(f, a...) printk("viafb: " f, ## a)

#define Kb      (1024)
#define Mb      (Kb*Kb)

#define CLE266          0x3122

struct fb_flip {
    u32 device;
    u32 field;
    u32 count;
    u32 offset[6];
};

#define VIAFB_FLIP_GRAPHICS 0
#define VIAFB_FLIP_V1 1
#define VIAFB_FLIP_V3 2
#define VIAFB_FLIP_SPIC 3
#define VIAFB_FLIP_NOP 255

#define VIAFB_WAIT_ANY 0
#define VIAFB_WAIT_TOPFIELD 1
#define VIAFB_WAIT_BOTTOMFIELD 2
#define VIAFB_WAIT_FLIP 3

/* 
 * This structure defines the hardware state of the graphics card. Normally
 * you place this in a header file in linux/include/video. This file usually
 * also includes register information. That allows other driver subsystems
 * and userland applications the ability to use the same header file to 
 * avoid duplicate work and easy porting of software. 
 */
struct via_par {
    unsigned int io_base_phy;
    unsigned int io_base_virt;
    unsigned int io_size;
    unsigned int mem_base_phy;
    unsigned long mem_base_virt;
    unsigned int mem_size;          // size of the total videomemory
    unsigned int video_mem_size;    // size of the available videomemory 
    unsigned int vq_start;
    unsigned int vq_end;
    unsigned int vq_size;
    unsigned int cur_start;
    unsigned int cur_end;
    unsigned int cur_size;
    unsigned int noaccel;
    unsigned int memsize;
    unsigned int refresh;
    unsigned int TVon;
    unsigned int TVindex;
    unsigned int TVtype;            // 0:none  1:NTSC  2:PAL
    unsigned int TVoverscan;        // 0:VIA_TVNORMAL  1:VIA_TVOVER
    unsigned int TVEncoder;
    
    int irq_num;    // Vertical blanking IRQ number
    int irq_cnt;    // Vertical blanking IRQ counter

    wait_queue_head_t wq;
    
    int vclk;       // in MHz
    
    struct fb_flip flip;
    int flip_pending;
};

#define VIASETREG(mmio, addr, data)   *(volatile u32 *)((mmio) +(addr)) = (data)
#define VIAGETREG(mmio, addr)         *(volatile u32 *)((mmio) +(addr))

#endif
